from functools import total_ordering

from ..util.xml import ET


@total_ordering
class File:
    """represent a file, including its metadata"""

    def __init__(self, name, md5, size, mtime, skipped=False):
        self.name = name
        self.md5 = md5
        self.size = size
        self.mtime = mtime
        self.skipped = skipped

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other
        self_data = (self.name, self.md5, self.size, self.mtime, self.skipped)
        other_data = (other.name, other.md5, other.size, other.mtime, other.skipped)
        return self_data == other_data

    def __lt__(self, other):
        self_data = (self.name, self.md5, self.size, self.mtime, self.skipped)
        other_data = (other.name, other.md5, other.size, other.mtime, other.skipped)
        return self_data < other_data

    @classmethod
    def from_xml_node(cls, node):
        assert node.tag == "entry"
        kwargs = {
            "name": node.get("name"),
            "md5": node.get("md5"),
            "size": int(node.get("size")),
            "mtime": int(node.get("mtime")),
            "skipped": "skipped" in node.attrib,
        }
        return cls(**kwargs)

    def to_xml_node(self, parent_node):
        attributes = {
            "name": self.name,
            "md5": self.md5,
            "size": str(int(self.size)),
            "mtime": str(int(self.mtime)),
        }
        if self.skipped:
            attributes["skipped"] = "true"
        new_node = ET.SubElement(parent_node, "entry", attributes)
        return new_node



import os
import unittest
from xml.etree import ElementTree as ET

import osc.core
import osc.oscerr
from osc.util.xml import xml_fromstring

from .common import OscTestCase


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'request_fixtures')


def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(TestRequest)


class TestRequest(OscTestCase):
    def _get_fixtures_dir(self):
        return FIXTURES_DIR

    def setUp(self):
        super().setUp(copytree=False)

    def test_createsr(self):
        """create a simple submitrequest"""
        r = osc.core.Request()
        r.add_action('submit', src_project='foo', src_package='bar', src_rev='42',
                     tgt_project='foobar', tgt_package='bar')
        self.assertEqual(r.actions[0].type, 'submit')
        self.assertEqual(r.actions[0].src_project, 'foo')
        self.assertEqual(r.actions[0].src_package, 'bar')
        self.assertEqual(r.actions[0].src_rev, '42')
        self.assertEqual(r.actions[0].tgt_project, 'foobar')
        self.assertEqual(r.actions[0].tgt_package, 'bar')
        self.assertTrue(r.actions[0].opt_sourceupdate is None)
        self.assertTrue(r.actions[0].opt_updatelink is None)
        self.assertTrue(len(r.statehistory) == 0)
        self.assertTrue(len(r.reviews) == 0)
        self.assertRaises(AttributeError, getattr, r.actions[0], 'doesnotexist')
        exp = """<request>
  <action type="submit">
    <source package="bar" project="foo" rev="42" />
    <target package="bar" project="foobar" />
  </action>
</request>"""
        self.assertXMLEqual(exp, r.to_str())

    def test_createsr_with_option(self):
        """create a submitrequest with option"""
        """create a simple submitrequest"""
        r = osc.core.Request()
        r.add_action('submit', src_project='foo', src_package='bar',
                     tgt_project='foobar', tgt_package='bar', opt_sourceupdate='cleanup', opt_updatelink='1')
        self.assertEqual(r.actions[0].type, 'submit')
        self.assertEqual(r.actions[0].src_project, 'foo')
        self.assertEqual(r.actions[0].src_package, 'bar')
        self.assertEqual(r.actions[0].tgt_project, 'foobar')
        self.assertEqual(r.actions[0].tgt_package, 'bar')
        self.assertEqual(r.actions[0].opt_sourceupdate, 'cleanup')
        self.assertEqual(r.actions[0].opt_updatelink, '1')
        self.assertTrue(r.actions[0].src_rev is None)
        self.assertTrue(len(r.statehistory) == 0)
        self.assertTrue(len(r.reviews) == 0)
        self.assertRaises(AttributeError, getattr, r.actions[0], 'doesnotexist')
        exp = """<request>
  <action type="submit">
    <source package="bar" project="foo" />
    <target package="bar" project="foobar" />
    <options>
      <sourceupdate>cleanup</sourceupdate>
      <updatelink>1</updatelink>
    </options>
  </action>
</request>"""
        self.assertXMLEqual(exp, r.to_str())

    def test_createsr_missing_tgt_package(self):
        """create a submitrequest with missing target package"""
        r = osc.core.Request()
        r.add_action('submit', src_project='foo', src_package='bar',
                     tgt_project='foobar')
        self.assertEqual(r.actions[0].type, 'submit')
        self.assertEqual(r.actions[0].src_project, 'foo')
        self.assertEqual(r.actions[0].src_package, 'bar')
        self.assertEqual(r.actions[0].tgt_project, 'foobar')
        self.assertTrue(len(r.statehistory) == 0)
        self.assertTrue(len(r.reviews) == 0)
        self.assertTrue(r.actions[0].tgt_package is None)
        self.assertRaises(AttributeError, getattr, r.actions[0], 'doesnotexist')
        exp = """<request>
  <action type="submit">
    <source package="bar" project="foo" />
    <target project="foobar" />
  </action>
</request>"""
        self.assertXMLEqual(exp, r.to_str())

    def test_createsr_invalid_argument(self):
        """create a submitrequest with invalid action argument"""
        r = osc.core.Request()
        self.assertRaises(osc.oscerr.WrongArgs, r.add_action, 'submit', src_project='foo', src_invalid='bar')

    def test_create_request_invalid_type(self):
        """create a request with an invalid action type"""
        r = osc.core.Request()
        self.assertRaises(osc.oscerr.WrongArgs, r.add_action, 'invalid', foo='bar')

    def test_create_add_role_person(self):
        """create an add_role request (person element)"""
        r = osc.core.Request()
        r.add_action('add_role', tgt_project='foo', tgt_package='bar', person_name='user', person_role='reader')
        self.assertEqual(r.actions[0].type, 'add_role')
        self.assertEqual(r.actions[0].tgt_project, 'foo')
        self.assertEqual(r.actions[0].tgt_package, 'bar')
        self.assertEqual(r.actions[0].person_name, 'user')
        self.assertEqual(r.actions[0].person_role, 'reader')
        self.assertTrue(r.actions[0].group_name is None)
        self.assertTrue(r.actions[0].group_role is None)
        exp = """<request>
  <action type="add_role">
    <target package="bar" project="foo" />
    <person name="user" role="reader" />
  </action>
</request>"""
        self.assertXMLEqual(exp, r.to_str())

    def test_create_add_role_group(self):
        """create an add_role request (group element)"""
        r = osc.core.Request()
        r.add_action('add_role', tgt_project='foo', tgt_package='bar', group_name='group', group_role='reviewer')
        self.assertEqual(r.actions[0].type, 'add_role')
        self.assertEqual(r.actions[0].tgt_project, 'foo')
        self.assertEqual(r.actions[0].tgt_package, 'bar')
        self.assertEqual(r.actions[0].group_name, 'group')
        self.assertEqual(r.actions[0].group_role, 'reviewer')
        self.assertTrue(r.actions[0].person_name is None)
        self.assertTrue(r.actions[0].person_role is None)
        self.assertTrue(len(r.statehistory) == 0)
        self.assertTrue(len(r.reviews) == 0)
        exp = """<request>
  <action type="add_role">
    <target package="bar" project="foo" />
    <group name="group" role="reviewer" />
  </action>
</request>"""
        self.assertXMLEqual(exp, r.to_str())

    def test_create_add_role_person_group(self):
        """create an add_role request (person+group element)"""
        r = osc.core.Request()
        r.add_action('add_role', tgt_project='foo', tgt_package='bar', person_name='user', person_role='reader',
                     group_name='group', group_role='reviewer')
        self.assertEqual(r.actions[0].type, 'add_role')
        self.assertEqual(r.actions[0].tgt_project, 'foo')
        self.assertEqual(r.actions[0].tgt_package, 'bar')
        self.assertEqual(r.actions[0].person_name, 'user')
        self.assertEqual(r.actions[0].person_role, 'reader')
        self.assertEqual(r.actions[0].group_name, 'group')
        self.assertEqual(r.actions[0].group_role, 'reviewer')
        self.assertTrue(len(r.statehistory) == 0)
        self.assertTrue(len(r.reviews) == 0)
        exp = """<request>
  <action type="add_role">
    <target package="bar" project="foo" />
    <person name="user" role="reader" />
    <group name="group" role="reviewer" />
  </action>
</request>"""
        self.assertXMLEqual(exp, r.to_str())

    def test_create_set_bugowner_project(self):
        """create a set_bugowner request for a project"""
        r = osc.core.Request()
        r.add_action('set_bugowner', tgt_project='foobar', person_name='buguser')
        self.assertEqual(r.actions[0].type, 'set_bugowner')
        self.assertEqual(r.actions[0].tgt_project, 'foobar')
        self.assertEqual(r.actions[0].person_name, 'buguser')
        self.assertTrue(r.actions[0].tgt_package is None)
        self.assertTrue(len(r.statehistory) == 0)
        self.assertTrue(len(r.reviews) == 0)
        exp = """<request>
  <action type="set_bugowner">
    <target project="foobar" />
    <person name="buguser" />
  </action>
</request>"""
        self.assertXMLEqual(exp, r.to_str())

    def test_create_set_bugowner_package(self):
        """create a set_bugowner request for a package"""
        r = osc.core.Request()
        r.add_action('set_bugowner', tgt_project='foobar', tgt_package='baz', person_name='buguser')
        self.assertEqual(r.actions[0].type, 'set_bugowner')
        self.assertEqual(r.actions[0].tgt_project, 'foobar')
        self.assertEqual(r.actions[0].tgt_package, 'baz')
        self.assertEqual(r.actions[0].person_name, 'buguser')
        self.assertTrue(len(r.statehistory) == 0)
        self.assertTrue(len(r.reviews) == 0)
        exp = """<request>
  <action type="set_bugowner">
    <target package="baz" project="foobar" />
    <person name="buguser" />
  </action>
</request>"""
        self.assertXMLEqual(exp, r.to_str())

    def test_create_delete_project(self):
        """create a delete request for a project"""
        r = osc.core.Request()
        r.add_action('delete', tgt_project='foo')
        self.assertEqual(r.actions[0].type, 'delete')
        self.assertEqual(r.actions[0].tgt_project, 'foo')
        self.assertTrue(r.actions[0].tgt_package is None)
        self.assertTrue(len(r.statehistory) == 0)
        self.assertTrue(len(r.reviews) == 0)
        exp = """<request>
  <action type="delete">
    <target project="foo" />
  </action>
</request>"""
        self.assertXMLEqual(exp, r.to_str())

    def test_create_delete_package(self):
        """create a delete request for a package"""
        r = osc.core.Request()
        r.add_action('delete', tgt_project='foo', tgt_package='deleteme')
        self.assertEqual(r.actions[0].type, 'delete')
        self.assertEqual(r.actions[0].tgt_project, 'foo')
        self.assertEqual(r.actions[0].tgt_package, 'deleteme')
        self.assertTrue(len(r.statehistory) == 0)
        self.assertTrue(len(r.reviews) == 0)
        exp = """<request>
  <action type="delete">
    <target package="deleteme" project="foo" />
  </action>
</request>"""
        self.assertXMLEqual(exp, r.to_str())

    def test_create_change_devel(self):
        """create a change devel request"""
        r = osc.core.Request()
        r.add_action('change_devel', src_project='foo', src_package='bar', tgt_project='devprj', tgt_package='devpkg')
        self.assertEqual(r.actions[0].type, 'change_devel')
        self.assertEqual(r.actions[0].src_project, 'foo')
        self.assertEqual(r.actions[0].src_package, 'bar')
        self.assertEqual(r.actions[0].tgt_project, 'devprj')
        self.assertEqual(r.actions[0].tgt_package, 'devpkg')
        self.assertTrue(len(r.statehistory) == 0)
        self.assertTrue(len(r.reviews) == 0)
        exp = """<request>
  <action type="change_devel">
    <source package="bar" project="foo" />
    <target package="devpkg" project="devprj" />
  </action>
</request>"""
        self.assertXMLEqual(exp, r.to_str())

    def test_action_from_xml1(self):
        """create action from xml"""
        xml = """<action type="add_role">
  <target package="bar" project="foo" />
  <person name="user" role="reader" />
  <group name="group" role="reviewer" />
</action>"""
        action = osc.core.Action.from_xml(xml_fromstring(xml))
        self.assertEqual(action.type, 'add_role')
        self.assertEqual(action.tgt_project, 'foo')
        self.assertEqual(action.tgt_package, 'bar')
        self.assertEqual(action.person_name, 'user')
        self.assertEqual(action.person_role, 'reader')
        self.assertEqual(action.group_name, 'group')
        self.assertEqual(action.group_role, 'reviewer')
        self.assertXMLEqual(xml, action.to_str())

    def test_action_from_xml2(self):
        """create action from xml"""
        xml = """<action type="submit">
  <source package="bar" project="foo" />
  <target package="bar" project="foobar" />
  <options>
    <sourceupdate>cleanup</sourceupdate>
    <updatelink>1</updatelink>
  </options>
</action>"""
        action = osc.core.Action.from_xml(xml_fromstring(xml))
        self.assertEqual(action.type, 'submit')
        self.assertEqual(action.src_project, 'foo')
        self.assertEqual(action.src_package, 'bar')
        self.assertEqual(action.tgt_project, 'foobar')
        self.assertEqual(action.tgt_package, 'bar')
        self.assertEqual(action.opt_sourceupdate, 'cleanup')
        self.assertEqual(action.opt_updatelink, '1')
        self.assertTrue(action.src_rev is None)
        self.assertXMLEqual(xml, action.to_str())

    def test_action_from_xml3(self):
        """create action from xml (with acceptinfo element)"""
        xml = """<action type="submit">
  <source package="bar" project="testprj" />
  <target package="baz" project="foobar" />
  <acceptinfo rev="5" srcmd5="aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa" xsrcmd5="ffffffffffffffffffffffffffffffff" />
</action>"""
        action = osc.core.Action.from_xml(xml_fromstring(xml))
        self.assertEqual(action.type, 'submit')
        self.assertEqual(action.src_project, 'testprj')
        self.assertEqual(action.src_package, 'bar')
        self.assertEqual(action.tgt_project, 'foobar')
        self.assertEqual(action.tgt_package, 'baz')
        self.assertTrue(action.opt_sourceupdate is None)
        self.assertTrue(action.opt_updatelink is None)
        self.assertTrue(action.src_rev is None)
        self.assertEqual(action.acceptinfo_rev, '5')
        self.assertEqual(action.acceptinfo_srcmd5, 'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa')
        self.assertEqual(action.acceptinfo_xsrcmd5, 'ffffffffffffffffffffffffffffffff')
        self.assertTrue(action.acceptinfo_osrcmd5 is None)
        self.assertTrue(action.acceptinfo_oxsrcmd5 is None)
        self.assertXMLEqual(xml, action.to_str())

    def test_action_from_xml_unknown_type(self):
        """try to create action from xml with unknown type"""
        xml = '<action type="foo"><source package="bar" project="foo" /></action>'
        self.assertRaises(osc.oscerr.WrongArgs, osc.core.Action.from_xml, xml_fromstring(xml))

    def test_read_request1(self):
        """read in a request"""
        xml = self._get_fixture('test_read_request1.xml')
        r = osc.core.Request()
        r.read(xml_fromstring(xml))
        self.assertEqual(r.reqid, '42')
        self.assertEqual(r.actions[0].type, 'submit')
        self.assertEqual(r.actions[0].src_project, 'foo')
        self.assertEqual(r.actions[0].src_package, 'bar')
        self.assertEqual(r.actions[0].src_rev, '1')
        self.assertEqual(r.actions[0].tgt_project, 'foobar')
        self.assertEqual(r.actions[0].tgt_package, 'bar')
        self.assertTrue(r.actions[0].opt_sourceupdate is None)
        self.assertTrue(r.actions[0].opt_updatelink is None)
        self.assertEqual(r.actions[1].type, 'delete')
        self.assertEqual(r.actions[1].tgt_project, 'deleteme')
        self.assertTrue(r.actions[1].tgt_package is None)
        self.assertEqual(r.state.name, 'accepted')
        self.assertEqual(r.state.when, '2010-12-27T01:36:29')
        self.assertEqual(r.state.who, 'user1')
        self.assertEqual(r.state.approver, None)
        self.assertEqual(r.state.comment, '')
        self.assertEqual(r.statehistory[0].when, '2010-12-13T13:02:03')
        self.assertEqual(r.statehistory[0].who, 'creator')
        self.assertEqual(r.statehistory[0].comment, 'foobar')
        self.assertEqual(r.title, 'title of the request')
        self.assertEqual(r.description, 'this is a\nvery long\ndescription')
        self.assertTrue(len(r.statehistory) == 1)
        self.assertTrue(len(r.reviews) == 0)
        self.assertXMLEqual(xml, r.to_str())

    def test_read_request2(self):
        """read in a request (with reviews)"""
        xml = self._get_fixture('test_read_request2.xml')
        r = osc.core.Request()
        r.read(xml_fromstring(xml))
        self.assertEqual(r.reqid, '123')
        self.assertEqual(r.actions[0].type, 'submit')
        self.assertEqual(r.actions[0].src_project, 'xyz')
        self.assertEqual(r.actions[0].src_package, 'abc')
        self.assertTrue(r.actions[0].src_rev is None)
        self.assertEqual(r.actions[0].opt_sourceupdate, 'cleanup')
        self.assertEqual(r.actions[0].opt_updatelink, '1')
        self.assertEqual(r.actions[1].type, 'add_role')
        self.assertEqual(r.actions[1].tgt_project, 'home:foo')
        self.assertEqual(r.actions[1].person_name, 'bar')
        self.assertEqual(r.actions[1].person_role, 'maintainer')
        self.assertEqual(r.actions[1].group_name, 'groupxyz')
        self.assertEqual(r.actions[1].group_role, 'reader')
        self.assertTrue(r.actions[1].tgt_package is None)
        self.assertEqual(r.state.name, 'review')
        self.assertEqual(r.state.when, '2010-12-27T01:36:29')
        self.assertEqual(r.state.approver, 'someone')
        self.assertEqual(r.state.who, 'abc')
        self.assertEqual(r.state.comment, '')
        self.assertEqual(r.reviews[0].state, 'new')
        self.assertEqual(r.reviews[0].by_group, 'group1')
        self.assertEqual(r.reviews[0].when, '2010-12-28T00:11:22')
        self.assertEqual(r.reviews[0].who, 'abc')
        self.assertEqual(r.reviews[0].comment, 'review start')
        self.assertTrue(r.reviews[0].by_user is None)
        self.assertEqual(r.statehistory[0].when, '2010-12-11T00:00:00')
        self.assertEqual(r.statehistory[0].who, 'creator')
        self.assertEqual(r.statehistory[0].comment, '')
        self.assertEqual(r.creator, 'creator')
        self.assertTrue(len(r.statehistory) == 1)
        self.assertTrue(len(r.reviews) == 1)
        self.assertXMLEqual(xml, r.to_str())

    def test_read_request3(self):
        """read in a request (with an "empty" comment+description)"""
        xml = """<request creator="xyz" id="2">
  <action type="set_bugowner">
    <target project="foo" />
    <person name="buguser" />
  </action>
  <state name="new" when="2010-12-28T12:36:29" who="xyz">
    <comment></comment>
  </state>
  <description></description>
</request>"""
        r = osc.core.Request()
        r.read(xml_fromstring(xml))
        self.assertEqual(r.reqid, '2')
        self.assertEqual(r.actions[0].type, 'set_bugowner')
        self.assertEqual(r.actions[0].tgt_project, 'foo')
        self.assertEqual(r.actions[0].person_name, 'buguser')
        self.assertEqual(r.state.name, 'new')
        self.assertEqual(r.state.when, '2010-12-28T12:36:29')
        self.assertEqual(r.state.who, 'xyz')
        self.assertEqual(r.state.comment, '')
        self.assertEqual(r.description, '')
        self.assertTrue(len(r.statehistory) == 0)
        self.assertTrue(len(r.reviews) == 0)
        self.assertEqual(r.creator, 'xyz')
        exp = """<request creator="xyz" id="2">
  <action type="set_bugowner">
    <target project="foo" />
    <person name="buguser" />
  </action>
  <state name="new" when="2010-12-28T12:36:29" who="xyz" />
</request>"""

        self.assertXMLEqual(exp, r.to_str())

    def test_request_list_view1(self):
        """test the list_view method"""
        xml = self._get_fixture('test_request_list_view1.xml')
        exp = """\
    62  State:new        By:Admin        When:2010-12-29T14:57:25
        Created by: Admin
        set_bugowner:    buguser                                            foo
        add_role:        person: xyz as maintainer, group: group1 as reader foobar
        add_role:        person: abc as reviewer                            foo/bar
        change_devel:    foo/bar                                            developed in devprj/devpkg
        submit:          srcprj/srcpackage ->                               tgtprj/tgtpackage
        submit:          foo/bar ->                                         baz
        delete:                                                             deleteme
        delete:                                                             foo/bar\n"""
        r = osc.core.Request()
        r.read(xml_fromstring(xml))
        self.assertEqual(exp, r.list_view())

    def test_request_list_view2(self):
        """test the list_view method (with history elements and description)"""
        xml = self._get_fixture('test_request_list_view2.xml')
        r = osc.core.Request()
        r.read(xml_fromstring(xml))
        exp = """\
    21  State:accepted   By:foobar       When:2010-12-29T16:37:45
        Created by: foobar
        set_bugowner:    buguser                                            foo
        From: Created Request: user -> Review Approved: foobar
        Descr: This is a simple request with a lot of ... ... text and other
               stuff. This request also contains a description. This is useful
               to describe the request. blabla blabla\n"""
        self.assertEqual(exp, r.list_view())

    def test_request_str1(self):
        """test the __str__ method"""
        xml = self._get_fixture('test_request_str1.xml')
        r = osc.core.Request()
        r = osc.core.Request()
        r.read(xml_fromstring(xml))
        self.assertEqual(r.creator, 'creator')
        exp = """\
Request:    123
Created by: creator

Actions:
  submit:       xyz/abc(cleanup) -> foo ***update link***
  add_role:     person: bar as maintainer, group: groupxyz as reader home:foo

Message:
  just a samll description
  in order to describe this
  request - blablabla
  test.

State:
  review                                                        2010-12-27T01:36:29 abc
    | currently in review

Review:
  accepted   Group: group1                                      2010-12-29T00:11:22 abc
    | accepted
  new        Group: group1                                      2010-12-28T00:11:22 abc
    | review start

History:
  2010-12-12T00:00:00 creator                        revoked
  2010-12-11T00:00:00 creator                        new"""
        self.assertEqual(exp, str(r))

    def test_request_str2(self):
        """test the __str__ method"""
        xml = """\
<request creator="creator" id="98765">
  <action type="change_devel">
    <source project="devprj" package="devpkg" />
    <target project="foo" package="bar" />
  </action>
  <action type="delete">
    <target project="deleteme" />
  </action>
  <state name="new" when="2010-12-29T00:11:22" who="creator" />
</request>"""
        r = osc.core.Request()
        r.read(xml_fromstring(xml))
        self.assertEqual(r.creator, 'creator')
        exp = """\
Request:    98765
Created by: creator

Actions:
  change_devel: foo/bar developed in devprj/devpkg
  delete:       deleteme

Message:
  <no message>

State:
  new                                                           2010-12-29T00:11:22 creator"""
        self.assertEqual(exp, str(r))

    def test_legacy_request(self):
        """load old-style submitrequest"""
        xml = """\
<request creator="olduser" id="1234" type="submit">
  <submit>
    <source package="baz" project="foobar" />
    <target package="baz" project="foo" />
  </submit>
  <state name="new" when="2010-12-30T02:11:22" who="olduser" />
</request>"""
        r = osc.core.Request()
        r.read(xml_fromstring(xml))
        self.assertEqual(r.reqid, '1234')
        self.assertEqual(r.actions[0].type, 'submit')
        self.assertEqual(r.actions[0].src_project, 'foobar')
        self.assertEqual(r.actions[0].src_package, 'baz')
        self.assertEqual(r.actions[0].tgt_project, 'foo')
        self.assertEqual(r.actions[0].tgt_package, 'baz')
        self.assertTrue(r.actions[0].opt_sourceupdate is None)
        self.assertTrue(r.actions[0].opt_updatelink is None)
        self.assertEqual(r.state.name, 'new')
        self.assertEqual(r.state.when, '2010-12-30T02:11:22')
        self.assertEqual(r.state.who, 'olduser')
        self.assertEqual(r.state.comment, '')
        self.assertEqual(r.creator, 'olduser')
        exp = """\
<request creator="olduser" id="1234">
  <action type="submit">
    <source package="baz" project="foobar" />
    <target package="baz" project="foo" />
  </action>
  <state name="new" when="2010-12-30T02:11:22" who="olduser" />
</request>"""
        self.assertXMLEqual(exp, r.to_str())

    def test_get_actions(self):
        """test get_actions method"""
        xml = self._get_fixture('test_request_list_view1.xml')
        r = osc.core.Request()
        r.read(xml_fromstring(xml))
        sr_actions = r.get_actions('submit')
        self.assertTrue(len(sr_actions) == 2)
        for i in sr_actions:
            self.assertEqual(i.type, 'submit')
        self.assertTrue(len(r.get_actions('submit', 'delete', 'change_devel')) == 5)
        self.assertTrue(len(r.get_actions()) == 8)


if __name__ == '__main__':
    unittest.main()



import textwrap

from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class StatusData(XmlModel):
    XML_TAG = "data"

    class NameEnum(str, Enum):
        SOURCEPROJECT = "sourceproject"
        SOURCEPACKAGE = "sourcepackage"
        TARGETPROJECT = "targetproject"
        TARGETPACKAGE = "targetpackage"
        TOKEN = "token"
        ID = "id"

    name: NameEnum = Field(
        xml_attribute=True,
        description=textwrap.dedent(
            """
            Key.
            """
        ),
    )

    value: str = Field(
        xml_set_text=True,
        description=textwrap.dedent(
            """
            Value.
            """
        ),
    )



import osc.commandline


class PersonSearchCommand(osc.commandline.OscCommand):
    """
    Search a person (user)
    """

    name = "search"
    parent = "PersonCommand"

    def init_arguments(self):
        self.add_argument(
            "--login",
            help="Search by a login.",
        )
        self.add_argument(
            "--login-contains",
            metavar="SUBSTR",
            help="Search by a substring in a login.",
        )
        self.add_argument(
            "--realname-contains",
            metavar="SUBSTR",
            help="Search by a substring in a realname.",
        )
        self.add_argument(
            "--email",
            help="Search by an email address.",
        )
        self.add_argument(
            "--email-contains",
            metavar="SUBSTR",
            help="Search by a substring in an email address.",
        )

    def run(self, args):
        from .. import obs_api

        persons = obs_api.Person.search(
            args.apiurl,
            login=args.login,
            login__contains=args.login_contains,
            realname__contains=args.realname_contains,
            email=args.email,
            email__contains=args.email_contains,
        )

        for person in persons:
            print(person.to_human_readable_string())
            print()



import itertools
import os
import re
import sys
import tarfile
from io import BytesIO

from . import ar
from . import packagequery


HAVE_LZMA = True
try:
    import lzma
except ImportError:
    HAVE_LZMA = False

HAVE_ZSTD = True
try:
    # Note: zstd is not supporting stream compression types
    import zstandard
except ImportError:
    HAVE_ZSTD = False


class DebError(packagequery.PackageError):
    pass


class DebQuery(packagequery.PackageQuery, packagequery.PackageQueryResult):

    default_tags = (
        b'package', b'version', b'release', b'epoch',
        b'architecture', b'description', b'provides', b'depends',
        b'pre_depends', b'conflicts', b'breaks'
    )

    def __init__(self, fh):
        self._file = fh
        self._path = os.path.abspath(fh.name)
        self.filename_suffix = 'deb'
        self.fields = {}

    def read(self, all_tags=False, self_provides=True, *extra_tags):
        arfile = ar.Ar(fh=self._file)
        arfile.read()
        debbin = arfile.get_file(b'debian-binary')
        if debbin is None:
            raise DebError(self._path, 'no debian binary')
        if debbin.read() != b'2.0\n':
            raise DebError(self._path, 'invalid debian binary format')
        for open_func in [self._open_tar_gz,
                          self._open_tar_xz,
                          self._open_tar_zst,
                          self._open_tar]:
            tar = open_func(arfile)
            if tar is not None:
                break
        if tar is None:
            raise DebError(self._path, 'missing control.tar')
        try:
            name = './control'
            control = tar.extractfile(name)
        except KeyError:
            raise DebError(self._path,
                           'missing \'control\' file in control.tar')
        self._parse_control(control, all_tags, self_provides, *extra_tags)
        return self

    def _open_tar(self, arfile):
        control = arfile.get_file(b'control.tar')
        if not control:
            return None

        return tarfile.open(fileobj=control)

    def _open_tar_gz(self, arfile):
        control = arfile.get_file(b'control.tar.gz')
        if not control:
            return None

        return tarfile.open(fileobj=control)

    def _open_tar_xz(self, arfile):
        control = arfile.get_file(b'control.tar.xz')
        if not control:
            return None

        if not HAVE_LZMA:
            raise DebError(self._path, 'can\'t open control.tar.xz without python-lzma')
        decompressed = lzma.decompress(control.read())
        return tarfile.open(fileobj=BytesIO(decompressed))

    def _open_tar_zst(self, arfile):
        control = arfile.get_file(b'control.tar.zst')
        if not control:
            return None

        if not HAVE_ZSTD:
            raise DebError(self._path, 'can\'t open control.tar.zst without python-zstandard')
        with zstandard.ZstdDecompressor().stream_reader(BytesIO(control.read())) as reader:
            decompressed = reader.read()
        return tarfile.open(fileobj=BytesIO(decompressed))

    def _parse_control(self, control, all_tags=False, self_provides=True, *extra_tags):
        data = control.readline().strip()
        while data:
            field, val = re.split(br':\s*', data.strip(), 1)
            data = control.readline()
            while data and re.match(br'\s+', data):
                val += b'\n' + data.strip()
                data = control.readline().rstrip()
            field = field.replace(b'-', b'_').lower()
            if field in self.default_tags + extra_tags or all_tags:
                # a hyphen is not allowed in dict keys
                self.fields[field] = val
        versrel = self.fields[b'version'].rsplit(b'-', 1)
        if len(versrel) == 2:
            self.fields[b'version'] = versrel[0]
            self.fields[b'release'] = versrel[1]
        else:
            self.fields[b'release'] = None
        verep = self.fields[b'version'].split(b':', 1)
        if len(verep) == 2:
            self.fields[b'epoch'] = verep[0]
            self.fields[b'version'] = verep[1]
        else:
            self.fields[b'epoch'] = b'0'
        self.fields[b'provides'] = self._split_field_value(b'provides')
        self.fields[b'depends'] = self._split_field_value(b'depends')
        self.fields[b'pre_depends'] = self._split_field_value(b'pre_depends')
        self.fields[b'conflicts'] = self._split_field_value(b'conflicts')
        self.fields[b'breaks'] = self._split_field_value(b'breaks')
        self.fields[b'recommends'] = self._split_field_value(b'recommends')
        self.fields[b'suggests'] = self._split_field_value(b'suggests')
        self.fields[b'enhances'] = self._split_field_value(b'enhances')
        if self_provides:
            # add self provides entry
            self.fields[b'provides'].append(b'%s (= %s)' % (self.name(), b'-'.join(versrel)))

    def _split_field_value(self, field, delimeter=br',\s*'):
        return [i.strip()
                for i in re.split(delimeter, self.fields.get(field, b'')) if i]

    def vercmp(self, debq):
        res = packagequery.cmp(int(self.epoch()), int(debq.epoch()))
        if res != 0:
            return res
        res = DebQuery.debvercmp(self.version(), debq.version())
        if res != 0:
            return res
        res = DebQuery.debvercmp(self.release(), debq.release())
        return res

    def name(self):
        return self.fields[b'package']

    def version(self):
        return self.fields[b'version']

    def release(self):
        return self.fields[b'release']

    def epoch(self):
        return self.fields[b'epoch']

    def arch(self):
        return self.fields[b'architecture']

    def description(self):
        return self.fields[b'description']

    def path(self):
        return self._path

    def provides(self):
        return self.fields[b'provides']

    def requires(self):
        return self.fields[b'depends'] + self.fields[b'pre_depends']

    def conflicts(self):
        return self.fields[b'conflicts'] + self.fields[b'breaks']

    def obsoletes(self):
        return []

    def recommends(self):
        return self.fields[b'recommends']

    def suggests(self):
        return self.fields[b'suggests']

    def supplements(self):
        # a control file has no notion of "supplements"
        return []

    def enhances(self):
        return self.fields[b'enhances']

    def gettag(self, num):
        return self.fields.get(num, None)

    def canonname(self):
        return DebQuery.filename(self.name(), self.epoch(), self.version(), self.release(), self.arch())

    @staticmethod
    def query(filename, all_tags=False, *extra_tags):
        f = open(filename, 'rb')
        debq = DebQuery(f)
        debq.read(all_tags, *extra_tags)
        f.close()
        return debq

    @staticmethod
    def debvercmp(ver1, ver2):
        """
        implementation of dpkg's version comparison algorithm
        """
        # 32 is arbitrary - it is needed for the "longer digit string wins" handling
        # (found this nice approach in Build/Deb.pm (build package))
        ver1 = re.sub(br'(\d+)', lambda m: (32 * b'0' + m.group(1))[-32:], ver1)
        ver2 = re.sub(br'(\d+)', lambda m: (32 * b'0' + m.group(1))[-32:], ver2)
        vers = itertools.zip_longest(ver1, ver2, fillvalue=b'')
        for v1, v2 in vers:
            if v1 == v2:
                continue
            if not v1:
                # this makes the corresponding condition in the following
                # else part superfluous - keep the superfluous condition for
                # now (just to ease a (hopefully) upcoming refactoring (this
                # method really deserves a cleanup...))
                return -1
            if not v2:
                # see above
                return 1
            v1 = bytes(bytearray([v1]))
            v2 = bytes(bytearray([v2]))
            if (v1.isalpha() and v2.isalpha()) or (v1.isdigit() and v2.isdigit()):
                res = packagequery.cmp(v1, v2)
                if res != 0:
                    return res
            else:
                if v1 == b'~' or not v1:
                    return -1
                elif v2 == b'~' or not v2:
                    return 1
                ord1 = ord(v1)
                if not (v1.isalpha() or v1.isdigit()):
                    ord1 += 256
                ord2 = ord(v2)
                if not (v2.isalpha() or v2.isdigit()):
                    ord2 += 256
                if ord1 > ord2:
                    return 1
                else:
                    return -1
        return 0

    @staticmethod
    def filename(name, epoch, version, release, arch):
        if release:
            return b'%s_%s-%s_%s.deb' % (name, version, release, arch)
        else:
            return b'%s_%s_%s.deb' % (name, version, arch)


if __name__ == '__main__':
    try:
        debq = DebQuery.query(sys.argv[1])
    except DebError as e:
        print(e.msg)
        sys.exit(2)
    print(debq.name(), debq.version(), debq.release(), debq.arch())
    print(debq.description())
    print('##########')
    print(b'\n'.join(debq.provides()))
    print('##########')
    print(b'\n'.join(debq.requires()))



# Copyright (C) 2018 SUSE Linux.  All rights reserved.
# This program is free software; it may be used, copied, modified
# and distributed under the terms of the GNU General Public Licence,
# either version 2, or (at your option) any later version.


import os
from urllib.request import HTTPError
from urllib.parse import urlparse
from urllib.parse import unquote
from urllib.error import URLError

try:
    from urllib3.exceptions import URLSchemeUnknown
except ImportError:
    class URLSchemeUnknown(Exception):
        pass

from .core import streamfile


class OscFileGrabber:
    def __init__(self, progress_obj=None):
        self.progress_obj = progress_obj

    def urlgrab(self, url, filename=None, text=None):
        if filename is None:
            parts = urlparse(url)
            filename = os.path.basename(unquote(parts[2]))
        with open(filename, 'wb') as f:
            for i in streamfile(url, progress_obj=self.progress_obj,
                                text=text):
                f.write(i)


class OscMirrorGroup:
    def __init__(self, grabber, mirrors):
        self._grabber = grabber
        self._mirrors = mirrors

    def urlgrab(self, url, filename=None, text=None):
        for mirror in self._mirrors:
            try:
                self._grabber.urlgrab(mirror, filename, text)
                return True
            except (HTTPError, URLError, URLSchemeUnknown, KeyError) as e:
                # urllib3 1.25.10 throws a KeyError: pool_key_constructor = self.key_fn_by_scheme[scheme]
                # try next mirror
                pass

        return False



import os
import subprocess
import urllib
from typing import Optional
from typing import Tuple


class Git:
    def __init__(self, workdir):
        self.abspath = os.path.abspath(workdir)

    def _run_git(self, args: list) -> str:
        return subprocess.check_output(["git"] + args, encoding="utf-8", cwd=self.abspath).strip()

    def init(self, *, quiet=True):
        cmd = ["init"]
        if quiet:
            cmd += ["-q"]
        self._run_git(cmd)

    # BRANCHES

    @property
    def current_branch(self) -> str:
        return self._run_git(["branch", "--show-current"])

    def get_branch_head(self, branch: str) -> str:
        return self._run_git(["rev-parse", branch])

    def switch(self, branch: str):
        self._run_git(["switch", branch])

    def fetch_pull_request(
        self,
        pull_number: int,
        *,
        remote: str = "origin",
        force: bool = False,
    ):
        """
        Fetch pull/$pull_number/head to pull/$pull_number branch
        """
        target_branch = f"pull/{pull_number}"
        cmd = ["fetch", remote, f"pull/{pull_number}/head:{target_branch}"]
        if force:
            cmd += [
                "--force",
                "--update-head-ok",
            ]
        self._run_git(cmd)
        return target_branch

    # CONFIG

    def set_config(self, key: str, value: str):
        self._run_git(["config", key, value])

    # REMOTES

    def get_remote_url(self, name: str = "origin") -> str:
        return self._run_git(["remote", "get-url", name])

    def add_remote(self, name: str, url: str):
        self._run_git(["remote", "add", name, url])

    def fetch(self, name: Optional[str] = None):
        if name:
            cmd = ["fetch", name]
        else:
            cmd = ["fetch", "--all"]
        self._run_git(cmd)

    def get_owner_repo(self, remote: str = "origin") -> Tuple[str, str]:
        remote_url = self.get_remote_url(name=remote)
        if "@" in remote_url:
            # ssh://gitea@example.com:owner/repo.git
            # ssh://gitea@example.com:22/owner/repo.git
            remote_url = remote_url.rsplit("@", 1)[-1]
        parsed_remote_url = urllib.parse.urlparse(remote_url)
        path = parsed_remote_url.path
        if path.endswith(".git"):
            path = path[:-4]
        owner, repo = path.strip("/").split("/")[-2:]
        return owner, repo



from . import api
from .. import oscerr


def get_configuration_value(apiurl, option):
    url_path = ["configuration"]
    url_query = {}
    root = api.get(apiurl, url_path, url_query)
    node = api.find_node(root, "configuration", option)
    if node is None or not node.text:
        raise oscerr.APIError(f"Couldn't get configuration option '{option}'")
    return node.text



import os
import unittest

import osc.core
import osc.oscerr

from .common import OscTestCase


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'project_package_status_fixtures')


def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(TestProjectStatus)


class TestProjectStatus(OscTestCase):
    def _get_fixtures_dir(self):
        return FIXTURES_DIR

    def test_simple(self):
        """get the status of a package with state ' '"""
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        exp_st = ' '
        st = prj.status('simple')
        self.assertEqual(exp_st, st)

    def test_added(self):
        """get the status of an added package"""
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        exp_st = 'A'
        st = prj.status('added')
        self.assertEqual(exp_st, st)

    def test_deleted(self):
        """get the status of a deleted package"""
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        exp_st = 'D'
        st = prj.status('deleted')
        self.assertEqual(exp_st, st)

    def test_added_deleted(self):
        """
        get the status of a package which was added and deleted
        afterwards (with a non osc command)
        """
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        exp_st = '!'
        st = prj.status('added_deleted')
        self.assertEqual(exp_st, st)

    def test_missing(self):
        """
        get the status of a package with state " "
        which was removed by a non osc command
        """
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        exp_st = '!'
        st = prj.status('missing')
        self.assertEqual(exp_st, st)

    def test_deleted_deleted(self):
        """
        get the status of a package which was deleted (with an
        osc command) and afterwards the package directory was
        deleted with a non osc command
        """
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        exp_st = 'D'
        st = prj.status('deleted_deleted')
        self.assertEqual(exp_st, st)

    def test_unversioned_exists(self):
        """get the status of an unversioned package"""
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        exp_st = '?'
        st = prj.status('excluded')
        self.assertEqual(exp_st, st)

    def test_unversioned_nonexistent(self):
        """get the status of an unversioned, nonexistent package"""
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        self.assertRaises(osc.oscerr.OscIOError, prj.status, 'doesnotexist')

    def test_get_status(self):
        """get the status of the complete project"""
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        exp_st = [(' ', 'conflict'), (' ', 'simple'), ('A', 'added'), ('D', 'deleted'),
                  ('!', 'missing'), ('!', 'added_deleted'), ('D', 'deleted_deleted'), ('?', 'excluded')]
        st = prj.get_status()
        self.assertEqual(exp_st, st)

    def test_get_status_excl(self):
        """get the status of the complete project (exclude some states)"""
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        exp_st = [('A', 'added'), ('!', 'missing'), ('!', 'added_deleted')]
        st = prj.get_status('D', ' ', '?')
        self.assertEqual(exp_st, st)

    def test_get_pacobj_simple(self):
        """package exists"""
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        p = prj.get_pacobj('simple')
        self.assertTrue(isinstance(p, osc.core.Package))
        self.assertEqual(p.name, 'simple')

    def test_get_pacobj_added(self):
        """package has state 'A', also test pac_kwargs"""
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        p = prj.get_pacobj('added', progress_obj={})
        self.assertTrue(isinstance(p, osc.core.Package))
        self.assertEqual(p.name, 'added')
        self.assertEqual(p.progress_obj, {})

    def test_get_pacobj_deleted(self):
        """package has state 'D' and exists, also test pac_args"""
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        p = prj.get_pacobj('deleted', {})
        self.assertTrue(isinstance(p, osc.core.Package))
        self.assertEqual(p.name, 'deleted')
        self.assertEqual(p.progress_obj, {})

    def test_get_pacobj_missing(self):
        """package is missing"""
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        p = prj.get_pacobj('missing')
        self.assertTrue(isinstance(p, type(None)))

    def test_get_pacobj_deleted_deleted(self):
        """package has state 'D' and does not exist"""
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        p = prj.get_pacobj('deleted_deleted')
        self.assertTrue(isinstance(p, type(None)))

    def test_get_pacobj_unversioned(self):
        """package/dir has state '?'"""
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        p = prj.get_pacobj('excluded')
        self.assertTrue(isinstance(p, type(None)))

    def test_get_pacobj_nonexistent(self):
        """package/dir does not exist"""
        self._change_to_pkg('.')
        prj = osc.core.Project('.', getPackageList=False)
        p = prj.get_pacobj('doesnotexist')
        self.assertTrue(isinstance(p, type(None)))


if __name__ == '__main__':
    unittest.main()



import argparse
import copy
import importlib
import inspect
import os
import pkgutil
import sys
import textwrap
from typing import List

from . import cmdln


# python3.6 requires reading sys.real_prefix to detect virtualenv
IN_VENV = getattr(sys, "real_prefix", sys.base_prefix) != sys.prefix


class OscArgumentParser(argparse.ArgumentParser):
    def _get_formatter(self):
        # cache formatter to speed things a little bit up
        if not hasattr(self, "_formatter"):
            self._formatter = self.formatter_class(prog=self.prog)
        return self._formatter

    def add_argument(self, *args, **kwargs):
        # remember added arguments so we can add them to subcommands easily
        if not hasattr(self, "_added_arguments"):
            self._added_arguments = []
        self._added_arguments.append((args, kwargs))
        super().add_argument(*args, **kwargs)


class Command:
    #: Name of the command as used in the argument parser.
    name: str = None

    #: Optional aliases to the command.
    aliases: List[str] = []

    #: Whether the command is hidden from help.
    #: Defaults to ``False``.
    hidden: bool = False

    #: Name of the parent command class.
    #: Can be prefixed if the parent comes from a different location,
    #: for example ``osc.commands.<ClassName>`` when extending osc command with a plugin.
    #: See ``OscMainCommand.MODULES`` for available prefixes.
    parent: str = None

    def __init__(self, full_name, parent=None):
        self.full_name = full_name
        self.parent = parent
        self.subparsers = None

        if not self.name:
            raise ValueError(f"Command '{self.full_name}' has no 'name' set")

        if parent:
            self.parser = self.parent.subparsers.add_parser(
                self.name,
                aliases=self.aliases,
                help=self.get_help(),
                description=self.get_description(),
                formatter_class=cmdln.HelpFormatter,
                conflict_handler="resolve",
                prog=f"{self.main_command.name} [global opts] {self.name}",
            )
            self.parser.set_defaults(_selected_command=self)
        else:
            self.parser = OscArgumentParser(
                description=self.get_description(),
                formatter_class=cmdln.HelpFormatter,
                usage="%(prog)s [global opts] <command> [--help] [opts] [args]",
            )

        if self.parent:
            for arg_args, arg_kwargs in self.parent.parser._added_arguments:
                if not arg_args:
                    continue
                if not arg_args[0].startswith("-"):
                    continue
                if "--help" in arg_args:
                    continue

                arg_kwargs = arg_kwargs.copy()
                arg_kwargs["help"] = argparse.SUPPRESS
                arg_kwargs["default"] = argparse.SUPPRESS
                self.parser.add_argument(*arg_args, **arg_kwargs)

        self.init_arguments()

    def __repr__(self):
        return f"<osc plugin {self.full_name} at {self.__hash__():#x}>"

    def get_help(self):
        """
        Return the help text of the command.
        The first line of the docstring is returned by default.
        """
        if self.hidden:
            return argparse.SUPPRESS

        if not self.__doc__:
            return ""

        help_lines = self.__doc__.strip().splitlines()

        if not help_lines:
            return ""

        return help_lines[0]

    def get_description(self):
        """
        Return the description of the command.
        The docstring without the first line is returned by default.
        """
        if not self.__doc__:
            return ""

        help_lines = self.__doc__.strip().splitlines()

        if not help_lines:
            return ""

        # skip the first line that contains help text
        help_lines.pop(0)

        # remove any leading empty lines
        while help_lines and not help_lines[0]:
            help_lines.pop(0)

        result = "\n".join(help_lines)
        result = textwrap.dedent(result)
        return result

    @property
    def main_command(self):
        """
        Return reference to the main command that represents the executable
        and contains the main instance of ArgumentParser.
        """
        if not self.parent:
            return self
        return self.parent.main_command

    def add_argument(self, *args, **kwargs):
        """
        Add a new argument to the command's argument parser.
        See `argparse <https://docs.python.org/3/library/argparse.html>`_ documentation for allowed parameters.
        """
        self.parser.add_argument(*args, **kwargs)

    def init_arguments(self):
        """
        Override to add arguments to the argument parser.

        .. note::
            Make sure you're adding arguments only by calling ``self.add_argument()``.
            Using ``self.parser.add_argument()`` directly is not recommended
            because it disables argument intermixing.
        """

    def run(self, args):
        """
        Override to implement the command functionality.

        .. note::
            ``args.positional_args`` is a list containing any unknown (unparsed) positional arguments.

        .. note::
            Consider moving any reusable code into a library,
            leaving the command-line code only a thin wrapper on top of it.

            If the code is generic enough, it should be added to osc directly.
            In such case don't hesitate to open an `issue <https://github.com/openSUSE/osc/issues>`_.
        """
        raise NotImplementedError()

    def register(self, command_class, command_full_name):
        if not self.subparsers:
            # instantiate subparsers on first use
            self.subparsers = self.parser.add_subparsers(dest="command", title="commands")

        # Check for parser conflicts.
        # This is how Python 3.11+ behaves by default.
        if command_class.name in self.subparsers._name_parser_map:
            raise argparse.ArgumentError(self.subparsers, f"conflicting subparser: {command_class.name}")
        for alias in command_class.aliases:
            if alias in self.subparsers._name_parser_map:
                raise argparse.ArgumentError(self.subparsers, f"conflicting subparser alias: {alias}")

        command = command_class(command_full_name, parent=self)
        return command


class MainCommand(Command):
    MODULES = ()

    def __init__(self):
        super().__init__(self.__class__.__name__)
        self.command_classes = {}
        self.download_progress = None

    def post_parse_args(self, args):
        pass

    def run(self, args):
        cmd = getattr(args, "_selected_command", None)
        if not cmd:
            self.parser.error("Please specify a command")
        self.post_parse_args(args)
        return cmd.run(args)

    def load_command(self, cls, module_prefix):
        mod_cls_name = f"{module_prefix}.{cls.__name__}"
        parent_name = getattr(cls, "parent", None)
        if parent_name:
            # allow relative references to classes in the the same module/directory
            if "." not in parent_name:
                parent_name = f"{module_prefix}.{parent_name}"
            try:
                parent = self.main_command.command_classes[parent_name]
            except KeyError:
                msg = f"Failed to load command class '{mod_cls_name}' because it references parent '{parent_name}' that doesn't exist"
                print(msg, file=sys.stderr)
                return None
            cmd = parent.register(cls, mod_cls_name)
        else:
            cmd = self.main_command.register(cls, mod_cls_name)

        cmd.full_name = mod_cls_name
        self.main_command.command_classes[mod_cls_name] = cmd
        return cmd

    def load_commands(self):
        if IN_VENV:
            from . import output  # pylint: disable=import-outside-toplevel
            output.print_msg("Running in virtual environment, skipping loading plugins installed outside the virtual environment.", print_to="debug")

        for module_prefix, module_path in self.MODULES:
            module_path = os.path.expanduser(module_path)

            # some plugins have their modules installed next to them instead of site-packages
            if module_path not in sys.path:
                sys.path.append(module_path)

            for loader, module_name, _ in pkgutil.iter_modules(path=[module_path]):
                full_name = f"{module_prefix}.{module_name}"
                spec = loader.find_spec(full_name)
                mod = importlib.util.module_from_spec(spec)
                try:
                    spec.loader.exec_module(mod)
                except Exception as e:  # pylint: disable=broad-except
                    msg = f"Failed to load commands from module '{full_name}': {e}"
                    print(msg, file=sys.stderr)
                    continue
                for name in dir(mod):
                    if name.startswith("_"):
                        continue
                    cls = getattr(mod, name)
                    if not inspect.isclass(cls):
                        continue
                    if not issubclass(cls, Command):
                        continue
                    if cls.__module__ != full_name:
                        # skip classes that weren't defined directly in the loaded plugin module
                        continue
                    self.load_command(cls, module_prefix)

    def parse_args(self, *args, **kwargs):
        namespace, unknown_args = self.parser.parse_known_args(*args, **kwargs)

        unrecognized = [i for i in unknown_args if i.startswith("-")]
        if unrecognized:
            self.parser.error(f"unrecognized arguments: " + " ".join(unrecognized))

        namespace.positional_args = list(unknown_args)
        return namespace



import os
import re
import shutil
import sys
import tempfile
import unittest

import osc.commandline
import osc.core
import osc.oscerr

from .common import GET, POST, OscTestCase, EXPECTED_REQUESTS


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'prdiff_fixtures')
UPSTREAM = 'some:project'
BRANCH = 'home:user:branches:' + UPSTREAM


def rdiff_url(pkg, oldprj, newprj):
    return 'http://localhost/source/%s/%s?unified=1&opackage=%s&oproject=%s&cmd=diff&expand=1&tarlimit=0&filelimit=0' % \
        (newprj, pkg, pkg, oldprj.replace(':', '%3A'))


def request_url(prj):
    return "http://localhost/request" + f"?view=collection&project={prj}&states=new,review".replace(":", "%3A").replace(",", "%2C")


def GET_PROJECT_PACKAGES(*projects):
    def decorator(test_method):
        # decorators get applied in the reversed order (bottom-up)
        for project in reversed(projects):
            test_method = GET(f'http://localhost/source/{project}', file=f'{project}/directory')(test_method)
        return test_method
    return decorator


def POST_RDIFF(oldprj, newprj):
    def decorator(test_method):
        # decorators get applied in the reversed order (bottom-up)
        test_method = POST(rdiff_url('common-three', oldprj, newprj), exp='', text='')(test_method)
        test_method = POST(rdiff_url('common-two', oldprj, newprj), exp='', file='common-two-diff')(test_method)
        test_method = POST(rdiff_url('common-one', oldprj, newprj), exp='', text='')(test_method)
        return test_method
    return decorator


def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(TestProjectDiff)


class TestProjectDiff(OscTestCase):
    diff_hdr = 'Index: %s\n==================================================================='

    def setUp(self, copytree=True):
        super().setUp(copytree=copytree)
        self.tmpdir_fixtures = tempfile.mkdtemp(prefix='osc_test')
        shutil.copytree(self._get_fixtures_dir(), os.path.join(self.tmpdir_fixtures, "fixtures"))

    def tearDown(self):
        try:
            shutil.rmtree(self.tmpdir_fixtures)
        except:
            pass
        super().tearDown()

    def _get_fixtures_dir(self):
        return FIXTURES_DIR

    def _change_to_tmpdir(self, *args):
        os.chdir(os.path.join(self.tmpdir, *args))

    def _run_prdiff(self, *args):
        """Runs osc prdiff, returning captured STDOUT as a string."""
        cli = osc.commandline.Osc()
        argv = ['osc', '--no-keyring', 'prdiff']
        argv.extend(args)
        cli.main(argv=argv)
        return sys.stdout.getvalue()

    def testPrdiffTooManyArgs(self):
        def runner():
            self._run_prdiff('one', 'two', 'superfluous-arg')
        self.assertRaises(osc.oscerr.WrongArgs, runner)

    @GET_PROJECT_PACKAGES(UPSTREAM, BRANCH)
    @POST_RDIFF(UPSTREAM, BRANCH)
    @POST(rdiff_url('only-in-new', UPSTREAM, BRANCH), exp='', text='')
    def testPrdiffZeroArgs(self):
        exp = """identical: common-one
differs:   common-two
identical: common-three
identical: only-in-new
"""

        def runner():
            self._run_prdiff()

        os.chdir('/tmp')
        self.assertRaises(osc.oscerr.WrongArgs, runner)

        self._change_to_tmpdir(self.tmpdir_fixtures, "fixtures", UPSTREAM)
        self.assertRaises(osc.oscerr.WrongArgs, runner)

        self._change_to_tmpdir(self.tmpdir_fixtures, "fixtures", BRANCH)
        out = self._run_prdiff()
        self.assertEqualMultiline(out, exp)

    @GET_PROJECT_PACKAGES(UPSTREAM, BRANCH)
    @POST_RDIFF(UPSTREAM, BRANCH)
    @POST(rdiff_url('only-in-new', UPSTREAM, BRANCH), exp='', text='')
    def testPrdiffOneArg(self):
        self._change_to_tmpdir()
        exp = """identical: common-one
differs:   common-two
identical: common-three
identical: only-in-new
"""
        out = self._run_prdiff('home:user:branches:some:project')
        self.assertEqualMultiline(out, exp)

    @GET_PROJECT_PACKAGES('old:prj', 'new:prj')
    @POST_RDIFF('old:prj', 'new:prj')
    def testPrdiffTwoArgs(self):
        self._change_to_tmpdir()
        exp = """identical: common-one
differs:   common-two
identical: common-three
"""
        out = self._run_prdiff('old:prj', 'new:prj')
        self.assertEqualMultiline(out, exp)

    @GET_PROJECT_PACKAGES('old:prj', 'new:prj')
    @POST_RDIFF('old:prj', 'new:prj')
    def testPrdiffOldOnly(self):
        self._change_to_tmpdir()
        exp = """identical: common-one
differs:   common-two
identical: common-three
old only:  only-in-old
"""
        out = self._run_prdiff('--show-not-in-new', 'old:prj', 'new:prj')
        self.assertEqualMultiline(out, exp)

    @GET_PROJECT_PACKAGES('old:prj', 'new:prj')
    @POST_RDIFF('old:prj', 'new:prj')
    def testPrdiffNewOnly(self):
        self._change_to_tmpdir()
        exp = """identical: common-one
differs:   common-two
identical: common-three
new only:  only-in-new
"""
        out = self._run_prdiff('--show-not-in-old', 'old:prj', 'new:prj')
        self.assertEqualMultiline(out, exp)

    @GET_PROJECT_PACKAGES('old:prj', 'new:prj')
    @POST_RDIFF('old:prj', 'new:prj')
    def testPrdiffDiffstat(self):
        self._change_to_tmpdir()
        exp = """identical: common-one
differs:   common-two

 common-two |    1 +
 1 file changed, 1 insertion(+)

identical: common-three
"""
        out = self._run_prdiff('--diffstat', 'old:prj', 'new:prj')
        self.assertEqualMultiline(out, exp)

    @GET_PROJECT_PACKAGES('old:prj', 'new:prj')
    @POST_RDIFF('old:prj', 'new:prj')
    def testPrdiffUnified(self):
        self._change_to_tmpdir()
        exp = """identical: common-one
differs:   common-two

Index: common-two
===================================================================
--- common-two\t2013-01-18 19:18:38.225983117 +0000
+++ common-two\t2013-01-18 19:19:27.882082325 +0000
@@ -1,4 +1,5 @@
 line one
 line two
 line three
+an extra line
 last line

identical: common-three
"""
        out = self._run_prdiff('--unified', 'old:prj', 'new:prj')
        self.assertEqualMultiline(out, exp)

    @GET_PROJECT_PACKAGES('old:prj', 'new:prj')
    @POST(rdiff_url('common-two', 'old:prj', 'new:prj'), exp='', file='common-two-diff')
    @POST(rdiff_url('common-three', 'old:prj', 'new:prj'), exp='', text='')
    def testPrdiffInclude(self):
        self._change_to_tmpdir()
        exp = """differs:   common-two
identical: common-three
"""
        out = self._run_prdiff('--include', 'common-t',
                               'old:prj', 'new:prj')
        self.assertEqualMultiline(out, exp)

    @GET_PROJECT_PACKAGES('old:prj', 'new:prj')
    @POST(rdiff_url('common-two', 'old:prj', 'new:prj'), exp='', file='common-two-diff')
    @POST(rdiff_url('common-three', 'old:prj', 'new:prj'), exp='', text='')
    def testPrdiffExclude(self):
        self._change_to_tmpdir()
        exp = """differs:   common-two
identical: common-three
"""
        out = self._run_prdiff('--exclude', 'one', 'old:prj', 'new:prj')
        self.assertEqualMultiline(out, exp)

    @GET_PROJECT_PACKAGES('old:prj', 'new:prj')
    @POST(rdiff_url('common-two', 'old:prj', 'new:prj'), exp='', file='common-two-diff')
    def testPrdiffIncludeExclude(self):
        self._change_to_tmpdir()
        exp = """differs:   common-two
"""
        out = self._run_prdiff('--include', 'common-t',
                               '--exclude', 'three',
                               'old:prj', 'new:prj')
        self.assertEqualMultiline(out, exp)

    @GET_PROJECT_PACKAGES(UPSTREAM, BRANCH)
    @GET(request_url(UPSTREAM), exp='', file='request')
    @POST(rdiff_url('common-one', UPSTREAM, BRANCH), exp='', text='')
    @POST(rdiff_url('common-two', UPSTREAM, BRANCH), exp='', file='common-two-diff')
    @POST(rdiff_url('common-three', UPSTREAM, BRANCH), exp='', file='common-two-diff')
    @POST(rdiff_url('only-in-new', UPSTREAM, BRANCH), exp='', text='')
    def testPrdiffRequestsMatching(self):
        self._change_to_tmpdir()
        exp = """identical: common-one
differs:   common-two

148023  State:new        By:user         When:2013-01-11T11:04:14
        Created by: creator
        submit:          home:user:branches:some:project/common-two@7 ->    some:project
        Descr: - Fix it to work - Improve support for something

differs:   common-three
identical: only-in-new
"""
        out = self._run_prdiff('--requests', UPSTREAM, BRANCH)
        self.assertEqualMultiline(out, exp)

    # Reverse the direction of the diff.

    @GET_PROJECT_PACKAGES(BRANCH, UPSTREAM)
    @GET(request_url(BRANCH), exp='', file='no-requests')
    @POST(rdiff_url('common-one', BRANCH, UPSTREAM), exp='', text='')
    @POST(rdiff_url('common-two', BRANCH, UPSTREAM), exp='', file='common-two-diff')
    @POST(rdiff_url('common-three', BRANCH, UPSTREAM), exp='', file='common-two-diff')
    @POST(rdiff_url('only-in-new', BRANCH, UPSTREAM), exp='', text='')
    def testPrdiffRequestsSwitched(self):
        self._change_to_tmpdir()
        exp = """identical: common-one
differs:   common-two
differs:   common-three
identical: only-in-new
"""
        out = self._run_prdiff('--requests', BRANCH, UPSTREAM)
        self.assertEqualMultiline(out, exp)


if __name__ == '__main__':
    unittest.main()



import os
import unittest

import osc.core
import osc.oscerr

from .common import OscTestCase


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'project_package_status_fixtures')


def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(TestPackageStatus)


class TestPackageStatus(OscTestCase):
    def _get_fixtures_dir(self):
        return FIXTURES_DIR

    def test_allfiles(self):
        """get the status of all files in the wc"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        exp_st = [('A', 'add'), ('?', 'exists'), ('D', 'foo'), ('!', 'merge'), ('R', 'missing'),
                  ('!', 'missing_added'), ('M', 'nochange'), ('S', 'skipped'), (' ', 'test')]
        st = p.get_status()
        self.assertEqual(exp_st, st)

    def test_todo(self):
        """
        get the status of some files in the wc.
        """
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.todo = ['test', 'missing_added', 'foo']
        exp_st = [('D', 'foo'), ('!', 'missing_added')]
        st = p.get_status(False, ' ')
        self.assertEqual(exp_st, st)

    def test_todo_noexcl(self):
        """ get the status of some files in the wc. """
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.todo = ['test', 'missing_added', 'foo']
        exp_st = [('D', 'foo'), ('!', 'missing_added'), (' ', 'test')]
        st = p.get_status()
        self.assertEqual(exp_st, st)

    def test_exclude_state(self):
        """get the status of all files in the wc but exclude some states"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        exp_st = [('A', 'add'), ('?', 'exists'), ('D', 'foo')]
        st = p.get_status(False, '!', 'S', ' ', 'M', 'R')
        self.assertEqual(exp_st, st)

    def test_nonexistent(self):
        """get the status of a non existent file"""
        self._change_to_pkg('simple')
        p = osc.core.Package('.')
        p.todo = ['doesnotexist']
        self.assertRaises(osc.oscerr.OscIOError, p.get_status)

    def test_conflict(self):
        """get status of the wc (one file in conflict state)"""
        self._change_to_pkg('conflict')
        p = osc.core.Package('.')
        exp_st = [('C', 'conflict'), ('?', 'exists'), (' ', 'test')]
        st = p.get_status()
        self.assertEqual(exp_st, st)

    def test_excluded(self):
        """get status of the wc (ignore excluded files); package has state ' '"""
        self._change_to_pkg('excluded')
        p = osc.core.Package('.')
        exp_st = [('?', 'exists'), ('M', 'modified')]
        st = p.get_status(False, ' ')
        self.assertEqual(exp_st, st)

    def test_noexcluded(self):
        """get status of the wc (include excluded files)"""
        self._change_to_pkg('excluded')
        p = osc.core.Package('.')
        exp_st = [('?', '_linkerror'), ('?', 'exists'), ('?', 'foo.orig'), ('M', 'modified'), (' ', 'test')]
        st = p.get_status(True)
        self.assertEqual(exp_st, st)


if __name__ == '__main__':
    unittest.main()



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class KeyinfoSslcert(XmlModel):
    XML_TAG = "sslcert"

    keyid: Optional[str] = Field(
        xml_attribute=True,
    )

    serial: Optional[str] = Field(
        xml_attribute=True,
    )

    issuer: Optional[str] = Field(
        xml_attribute=True,
    )

    subject: Optional[str] = Field(
        xml_attribute=True,
    )

    algo: Optional[str] = Field(
        xml_attribute=True,
    )

    keysize: Optional[str] = Field(
        xml_attribute=True,
    )

    begins: Optional[int] = Field(
        xml_attribute=True,
    )

    expires: Optional[int] = Field(
        xml_attribute=True,
    )

    fingerprint: Optional[str] = Field(
        xml_attribute=True,
    )

    value: str = Field(
        xml_set_text=True,
    )

    def get_begins_str(self) -> str:
        import datetime

        if self.begins is None:
            return ""

        return datetime.datetime.fromtimestamp(self.begins).strftime("%Y-%m-%d %H:%M:%S")

    def get_expires_str(self) -> str:
        import datetime

        if self.expires is None:
            return ""

        return datetime.datetime.fromtimestamp(self.expires).strftime("%Y-%m-%d %H:%M:%S")

    def to_human_readable_string(self) -> str:
        """
        Render the object as a human readable string.
        """
        from ..output import KeyValueTable
        table = KeyValueTable()
        table.add("Type", "SSL certificate")
        table.add("Subject", self.subject, color="bold")
        table.add("Key ID", self.keyid)
        table.add("Serial", self.serial)
        table.add("Issuer", self.issuer)
        table.add("Algorithm", self.algo)
        table.add("Key size", self.keysize)
        table.add("Begins", self.get_begins_str())
        table.add("Expires", self.get_expires_str())
        table.add("Fingerprint", self.fingerprint)
        return f"{table}\n{self.value}"



import unittest

from osc.util.helper import decode_it, decode_list


def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(TestResults)


class TestResults(unittest.TestCase):
    def testDecodeList(self):
        strlist = ['Test1', 'Test2', 'Test3']
        mixlist = ['Test1', b'Test2', 'Test3']
        byteslist = [b'Test1', b'Test2', b'Test3']

        out = decode_list(strlist)
        self.assertListEqual(out, strlist)

        out = decode_list(mixlist)
        self.assertListEqual(out, strlist)

        out = decode_list(byteslist)
        self.assertListEqual(out, strlist)

    def testDecodeIt(self):
        bytes_obj = b'Test the decoding'
        string_obj = 'Test the decoding'

        out = decode_it(bytes_obj)
        self.assertEqual(out, string_obj)

        out = decode_it(string_obj)
        self.assertEqual(out, string_obj)


if __name__ == '__main__':
    unittest.main()



"""
Functions that manipulate with XML.
"""

import io
import xml.sax.saxutils
from typing import Union
from xml.etree import ElementTree as ET


def xml_escape(string):
    """
    Escape the string so it's safe to use in XML and xpath.
    """
    entities = {
        '"': "&quot;",
        "'": "&apos;",
    }
    if isinstance(string, bytes):
        return xml.sax.saxutils.escape(string.decode("utf-8"), entities=entities).encode("utf-8")
    return xml.sax.saxutils.escape(string, entities=entities)


def xml_unescape(string):
    """
    Decode XML entities in the string.
    """
    entities = {
        "&quot;": '"',
        "&apos;": "'",
    }
    if isinstance(string, bytes):
        return xml.sax.saxutils.unescape(string.decode("utf-8"), entities=entities).encode("utf-8")
    return xml.sax.saxutils.unescape(string, entities=entities)


def xml_strip_text(node):
    """
    Recursively strip inner text in nodes:
    - if text contains only whitespaces
    - if node contains child nodes
    """
    if node.text and not node.text.strip():
        node.text = None
    elif len(node) != 0:
        node.text = None
    for child in node:
        xml_strip_text(child)


def xml_indent_compat(elem, level=0):
    """
    XML indentation code for python < 3.9.
    Source: http://effbot.org/zone/element-lib.htm#prettyprint
    """
    i = "\n" + level * "  "
    if isinstance(elem, ET.ElementTree):
        elem = elem.getroot()
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        for e in elem:
            xml_indent_compat(e, level + 1)
            if not e.tail or not e.tail.strip():
                e.tail = i + "  "
        if not e.tail or not e.tail.strip():
            e.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


def xml_indent(root):
    """
    Indent XML so it looks pretty after printing or saving to file.
    """
    if hasattr(ET, "indent"):
        # ElementTree supports indent() in Python 3.9 and newer
        xml_strip_text(root)
        ET.indent(root)
    else:
        xml_indent_compat(root)


def _extend_parser_error_msg(e: ET.ParseError, text: Union[str, bytes]):
    from ..output import tty

    y, x = e.position
    text = text.splitlines()[y-1][x-1:]

    if isinstance(text, bytes):
        text = text.decode("utf-8")

    new_text = ""
    for char in text:
        if char >= " ":
            new_text += char
            continue
        byte = ord(char)
        char = f"0x{byte:0>2X}"
        char = tty.colorize(char, "bg_red")
        new_text += char
    e.msg += ": " + new_text


def xml_fromstring(text: str):
    """
    xml.etree.ElementTree.fromstring() wrapper that extends error message in ParseError
    exceptions with a snippet of the broken XML.
    """
    try:
        return ET.fromstring(text)
    except ET.ParseError as e:
        _extend_parser_error_msg(e, text)
        raise


def xml_parse(source):
    """
    xml.etree.ElementTree.parse() wrapper that extends error message in ParseError
    exceptions with a snippet of the broken XML.
    """
    if isinstance(source, str):
        # source is a file name
        with open(source, "rb") as f:
            data = f.read()
    else:
        # source is an IO object
        data = source.read()

    if isinstance(data, bytes):
        f = io.BytesIO(data)
    else:
        f = io.StringIO(data)

    try:
        return ET.parse(f)
    except ET.ParseError as e:
        _extend_parser_error_msg(e, data)
        raise



import unittest

from osc.obs_api.scmsync_obsinfo import ScmsyncObsinfo


class TestScmsyncObsinfo(unittest.TestCase):
    def test_empty(self):
        self.assertRaises(TypeError, ScmsyncObsinfo.from_string, "")

    def test_mandatory(self):
        data = """
        mtime: 123
        commit: abcdef
        """
        info = ScmsyncObsinfo.from_string(data)
        self.assertEqual(info.mtime, 123)
        self.assertEqual(info.commit, "abcdef")
        self.assertEqual(info.url, None)

    def test_all(self):
        data = """
        mtime: 123
        commit: abcdef
        url: https://example.com
        revision: 1
        subdir: dirname
        projectscmsync: project
        """
        info = ScmsyncObsinfo.from_string(data)
        self.assertEqual(info.mtime, 123)
        self.assertEqual(info.commit, "abcdef")
        self.assertEqual(info.url, "https://example.com")
        self.assertEqual(info.revision, "1")
        self.assertEqual(info.subdir, "dirname")
        self.assertEqual(info.projectscmsync, "project")


if __name__ == "__main__":
    unittest.main()



import unittest

from osc.util.xpath import XPathQuery as Q


class TestQuery(unittest.TestCase):
    def test_noop(self):
        q = Q(name="foo")
        self.assertEqual(str(q), "@name='foo'")

    def test_not(self):
        q = Q(name__not="foo")
        self.assertEqual(str(q), "not(@name='foo')")

    def test_eq(self):
        q = Q(name__eq="foo")
        self.assertEqual(str(q), "@name='foo'")

    def test_not_eq(self):
        q = Q(name__not__eq="foo")
        self.assertEqual(str(q), "not(@name='foo')")

    def test_contains(self):
        q = Q(name__contains="foo")
        self.assertEqual(str(q), "contains(@name, 'foo')")

    def test_and(self):
        q1 = Q(name="foo")
        q2 = Q(name="bar")
        q = q1 & q2
        self.assertEqual(str(q), "@name='foo' and @name='bar'")

        q3 = Q(name="baz")
        q = q & q3
        self.assertEqual(str(q), "@name='foo' and @name='bar' and @name='baz'")

    def test_or(self):
        q1 = Q(name="foo")
        q2 = Q(name="bar")
        q = q1 | q2
        self.assertEqual(str(q), "@name='foo' or @name='bar'")

        q3 = Q(name="baz")
        q = q | q3
        self.assertEqual(str(q), "@name='foo' or @name='bar' or @name='baz'")

    def test_and_or(self):
        q1 = Q(name="foo")
        q2 = Q(name="bar")
        q = q1 & q2
        self.assertEqual(str(q), "@name='foo' and @name='bar'")

        q3 = Q(name="baz")
        q = q | q3
        self.assertEqual(str(q), "(@name='foo' and @name='bar') or @name='baz'")

        q4 = Q(name="xyz")
        q = q | q4
        self.assertEqual(str(q), "(@name='foo' and @name='bar') or @name='baz' or @name='xyz'")

    def test_or_and(self):
        q1 = Q(name="foo")
        q2 = Q(name="bar")
        q = q1 | q2
        self.assertEqual(str(q), "@name='foo' or @name='bar'")

        q3 = Q(name="baz")
        q = q & q3
        self.assertEqual(str(q), "(@name='foo' or @name='bar') and @name='baz'")

        q4 = Q(name="xyz")
        q = q & q4
        self.assertEqual(str(q), "(@name='foo' or @name='bar') and @name='baz' and @name='xyz'")

    def test_and_or_and(self):
        q1 = Q(name="foo")
        q2 = Q(name="bar")
        q3 = Q(name="baz")
        q4 = Q(name="xyz")
        q = (q1 & q2) | (q3 & q4)
        self.assertEqual(str(q), "(@name='foo' and @name='bar') or (@name='baz' and @name='xyz')")

    def test_or_and_or(self):
        q1 = Q(name="foo")
        q2 = Q(name="bar")
        q3 = Q(name="baz")
        q4 = Q(name="xyz")
        q = (q1 | q2) & (q3 | q4)
        self.assertEqual(str(q), "(@name='foo' or @name='bar') and (@name='baz' or @name='xyz')")

    def test_multiple_kwargs(self):
        q = Q(name1="foo", name2="bar")
        self.assertEqual(str(q), "@name1='foo' and @name2='bar'")

    def test_eq_list(self):
        q = Q(name=["foo", "bar", "baz"])
        self.assertEqual(str(q), "@name='foo' or @name='bar' or @name='baz'")

    def test_not_eq_list(self):
        q = Q(name__not=["foo", "bar", "baz"])
        self.assertEqual(str(q), "not(@name='foo') and not(@name='bar') and not(@name='baz')")

    def test_review_state(self):
        q = Q(state__name=["new"])
        self.assertEqual(str(q), "state[@name='new']")


if __name__ == "__main__":
    unittest.main()



import json
import sys

import osc.commandline_git


class ApiCommand(osc.commandline_git.GitObsCommand):
    """
    Make an arbitrary request to API
    """

    name = "api"

    def init_arguments(self):
        self.add_argument(
            "-X",
            "--method",
            choices=["GET", "HEAD", "POST", "PATCH", "PUT"],
            default="GET",
        )
        self.add_argument(
            "url",
        )
        self.add_argument("--data")

    def run(self, args):
        from osc import gitea_api
        from osc.output import tty

        self.print_gitea_settings()

        url = self.gitea_conn.makeurl(args.url)

        json_data = None
        if args.data:
            json_data = json.loads(args.data)

        response = self.gitea_conn.request(
            method=args.method, url=url, json_data=json_data
        )
        print(tty.colorize("Response:", "white,bold"), file=sys.stderr)
        if response.headers.get("Content-Type", "").startswith("application/json;"):
            print(
                json.dumps(
                    json.loads(response.data),
                    indent=4,
                    sort_keys=True,
                )
            )
        else:
            print(response.data.decode("utf-8"))



import osc.commandline_git


class SSHKeyListCommand(osc.commandline_git.GitObsCommand):
    """
    """

    name = "list"
    parent = "SSHKeyCommand"

    def init_arguments(self):
        pass

    def run(self, args):
        from osc import gitea_api

        self.print_gitea_settings()

        for i in gitea_api.SSHKey.list(self.gitea_conn).json():
            print(gitea_api.SSHKey.to_human_readable_string(i))
            print()



import os
import unittest

import osc.conf
import osc.core
import osc.oscerr

from .common import GET, OscTestCase


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'init_project_fixtures')


def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(TestInitProject)


class TestInitProject(OscTestCase):
    def _get_fixtures_dir(self):
        return FIXTURES_DIR

    def setUp(self):
        super().setUp(copytree=False)

    def test_simple(self):
        """initialize a project dir"""
        prj_dir = os.path.join(self.tmpdir, 'testprj')
        prj = osc.core.Project.init_project('http://localhost', prj_dir, 'testprj', getPackageList=False)
        self.assertTrue(isinstance(prj, osc.core.Project))
        storedir = os.path.join(prj_dir, osc.core.store)
        self._check_list(os.path.join(storedir, '_project'), 'testprj\n')
        self._check_list(os.path.join(storedir, '_apiurl'), 'http://localhost\n')
        self._check_list(os.path.join(storedir, '_packages'), '<project name="testprj" />')

    def test_dirExists(self):
        """initialize a project dir but the dir already exists"""
        prj_dir = os.path.join(self.tmpdir, 'testprj')
        os.mkdir(prj_dir)
        prj = osc.core.Project.init_project('http://localhost', prj_dir, 'testprj', getPackageList=False)
        self.assertTrue(isinstance(prj, osc.core.Project))
        storedir = os.path.join(prj_dir, osc.core.store)
        self._check_list(os.path.join(storedir, '_project'), 'testprj\n')
        self._check_list(os.path.join(storedir, '_apiurl'), 'http://localhost\n')
        self._check_list(os.path.join(storedir, '_packages'), '<project name="testprj" />')

    def test_storedirExists(self):
        """initialize a project dir but the storedir already exists"""
        prj_dir = os.path.join(self.tmpdir, 'testprj')
        os.mkdir(prj_dir)
        os.mkdir(os.path.join(prj_dir, osc.core.store))
        self.assertRaises(osc.oscerr.OscIOError, osc.core.Project.init_project, 'http://localhost', prj_dir, 'testprj')

    @GET('http://localhost/source/testprj', text='<directory count="0" />')
    def test_no_package_tracking(self):
        """initialize a project dir but disable package tracking; enable getPackageList=True;
        disable wc_check (because we didn't disable the package tracking before the Project class
        was imported therefore REQ_STOREFILES contains '_packages')
        """
        # disable package tracking
        osc.conf.config['do_package_tracking'] = False
        prj_dir = os.path.join(self.tmpdir, 'testprj')
        os.mkdir(prj_dir)
        prj = osc.core.Project.init_project('http://localhost', prj_dir, 'testprj', False, wc_check=False)
        self.assertTrue(isinstance(prj, osc.core.Project))
        storedir = os.path.join(prj_dir, osc.core.store)
        self._check_list(os.path.join(storedir, '_project'), 'testprj\n')
        self._check_list(os.path.join(storedir, '_apiurl'), 'http://localhost\n')
        self.assertFalse(os.path.exists(os.path.join(storedir, '_packages')))


if __name__ == '__main__':
    unittest.main()



import configparser
import unittest

from osc.OscConfigParser import OscConfigParser


class TestOscConfigParser(unittest.TestCase):
    def setUp(self):
        self.parser = OscConfigParser()
        self.parser.read_string("""
[general]
apiurl = http://localhost

[http://localhost]
credentials_mgr_class=
user=
pass=
""")

    def test_disabled_interpolation(self):
        # with interpolation on, this would raise
        # ValueError: invalid interpolation syntax in '%' at position 0
        self.parser.set("http://localhost", "pass", "%")

    def test_duplicate_section(self):
        conf = """
[general]

[http://localhost]

[http://localhost]
"""
        parser = OscConfigParser()
        self.assertRaises(configparser.DuplicateSectionError, parser.read_string, conf)

    def test_duplicate_option(self):
        conf = """
[general]

[http://localhost]
user=
user=
"""
        parser = OscConfigParser()
        self.assertRaises(configparser.DuplicateOptionError, parser.read_string, conf)


if __name__ == "__main__":
    unittest.main()



import os
import sys
import unittest

import osc.commandline

from .common import GET, OscTestCase


def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(TestResults)


class TestResults(OscTestCase):
    def setUp(self):
        super().setUp(copytree=False)

    def _get_fixtures_name(self):
        return 'results_fixtures'

    def _get_fixtures_dir(self):
        return os.path.join(os.path.dirname(__file__), self._get_fixtures_name())

    def _run_osc(self, *args):
        """Runs osc, returning captured STDOUT as a string."""
        cli = osc.commandline.Osc()
        argv = ['osc', '--no-keyring']
        argv.extend(args)
        cli.main(argv=argv)
        return sys.stdout.getvalue()

    @GET('http://localhost/build/testproject/_result', file='result.xml')
    def testPrjresults(self):
        out = self._run_osc('prjresults', '--xml', 'testproject')
        self.assertEqualMultiline(out, self._get_fixture('result.xml') + '\n')

    @GET('http://localhost/build/testproject/_result', file='result-dirty.xml')
    @GET('http://localhost/build/testproject/_result?oldstate=c57e2ee592dbbf26ebf19cc4f1bc1e83', file='result.xml')
    def testPrjresultsWatch(self):
        out = self._run_osc('prjresults', '--watch', '--xml', 'testproject')
        self.assertEqualMultiline(out, self._get_fixture('result-dirty.xml') + '\n' + self._get_fixture('result.xml') + '\n')

    @GET('http://localhost/build/testproject/_result?package=python-MarkupSafe&multibuild=1&locallink=1', file='result.xml')
    def testResults(self):
        out = self._run_osc('results', '--xml', 'testproject', 'python-MarkupSafe')
        self.assertEqualMultiline(out, self._get_fixture('result.xml'))

    @GET('http://localhost/build/testproject/_result?package=python-MarkupSafe&multibuild=1&locallink=1', file='result-dirty.xml')
    @GET('http://localhost/build/testproject/_result?package=python-MarkupSafe&oldstate=c57e2ee592dbbf26ebf19cc4f1bc1e83&multibuild=1&locallink=1', file='result.xml')
    def testResultsWatch(self):
        out = self._run_osc('results', '--watch', '--xml', 'testproject', 'python-MarkupSafe')
        self.assertEqualMultiline(out, self._get_fixture('result-dirty.xml') + self._get_fixture('result.xml'))


if __name__ == '__main__':
    unittest.main()



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class RequestHistory(XmlModel):
    XML_TAG = "history"

    who: str = Field(
        xml_attribute=True,
    )

    when: str = Field(
        xml_attribute=True,
    )

    description: str = Field(
    )

    comment: Optional[str] = Field(
    )



"""
A modern, lightweight alternative to cmdln.py from https://github.com/trentm/cmdln
"""


import argparse
import inspect
import sys
import textwrap


def option(*args, **kwargs):
    """
    Decorator to add an option to the optparser argument of a Cmdln subcommand.

    Example:
        class MyShell(cmdln.Cmdln):
            @cmdln.option("-f", "--force", help="force removal")
            def do_remove(self, subcmd, opts, *args):
                #...
    """
    def decorate(f):
        if not hasattr(f, "options"):
            f.options = []
        new_args = [i for i in args if i]
        f.options.insert(0, (new_args, kwargs))
        return f
    return decorate


def alias(*aliases):
    """
    Decorator to add aliases for Cmdln.do_* command handlers.

    Example:
        class MyShell(cmdln.Cmdln):
            @cmdln.alias("!", "sh")
            def do_shell(self, argv):
                #...implement 'shell' command
    """
    def decorate(f):
        if not hasattr(f, "aliases"):
            f.aliases = []
        f.aliases += aliases
        return f
    return decorate


def name(name):
    """
    Decorator to explicitly name a Cmdln subcommand.

    Example:
        class MyShell(cmdln.Cmdln):
            @cmdln.name("cmd-with-dashes")
            def do_cmd_with_dashes(self, subcmd, opts):
                #...
    """
    def decorate(f):
        f.name = name
        return f
    return decorate


def hide(value=True):
    """
    For obsolete calls, hide them in help listings.

    Example:
        class MyShell(cmdln.Cmdln):
            @cmdln.hide()
            def do_shell(self, argv):
                #...implement 'shell' command
    """
    def decorate(f):
        f.hidden = bool(value)
        return f
    return decorate


class HelpFormatter(argparse.RawDescriptionHelpFormatter):
    def _split_lines(self, text, width):
        # remove the leading and trailing whitespaces to avoid printing unwanted blank lines
        text = text.strip()

        result = []
        for line in text.splitlines():
            if not line.strip():
                # textwrap normally returns [] on a string that contains only whitespaces; we want [""] to print a blank line
                result.append("")
            else:
                result.extend(textwrap.wrap(line, width))
        return result

    def _format_action(self, action):
        if isinstance(action, argparse._SubParsersAction):
            parts = []
            subactions = action._get_subactions()
            subactions.sort(key=lambda x: x.metavar)
            for i in subactions:
                if i.help == argparse.SUPPRESS:
                    # don't display commands with suppressed help
                    continue
                if len(i.metavar) > 20:
                    parts.append("%*s%-21s" % (self._current_indent, "", i.metavar))
                    parts.append("%*s %s" % (self._current_indent + 21, "", i.help))
                else:
                    parts.append("%*s%-21s %s" % (self._current_indent, "", i.metavar, i.help))
            return "\n".join(parts)
        return super()._format_action(action)


class Cmdln:
    def get_argparser_usage(self):
        return "%(prog)s [global opts] <command> [--help] [opts] [args]"

    def get_subcommand_prog(self, subcommand):
        return f"{self.argparser.prog} [global opts] {subcommand}"

    def _remove_leading_spaces_from_text(self, text):
        lines = text.splitlines()
        lines = self._remove_leading_spaces_from_lines(lines)
        return "\n".join(lines)

    def _remove_leading_spaces_from_lines(self, lines):
        # compute the indentation (leading spaces) in the docstring
        leading_spaces = 0
        for line in lines:
            line_leading_spaces = len(line) - len(line.lstrip(' '))
            if leading_spaces == 0:
                leading_spaces = line_leading_spaces
            leading_spaces = min(leading_spaces, line_leading_spaces)
        # dedent the lines (remove leading spaces)
        lines = [line[leading_spaces:] for line in lines]
        return lines

    def create_argparser(self):
        """
        Create `.argparser` and `.subparsers`.
        Override this method to replace them with your own.
        """
        self.argparser = argparse.ArgumentParser(
            usage=self.get_argparser_usage(),
            description=self._remove_leading_spaces_from_text(self.__doc__),
            formatter_class=HelpFormatter,
        )
        self.subparsers = self.argparser.add_subparsers(
            title="commands",
            dest="command",
        )

        self.pre_argparse()
        self.add_global_options(self.argparser)

        # map command name to `do_*` function that runs the command
        self.cmd_map = {}

        # map aliases back to the command names
        self.alias_to_cmd_name_map = {}

        for attr in dir(self):
            if not attr.startswith("do_"):
                continue

            cmd_name = attr[3:]
            cmd_func = getattr(self, attr)

            # extract data from the function
            cmd_name = getattr(cmd_func, "name", cmd_name)
            options = getattr(cmd_func, "options", [])
            aliases = getattr(cmd_func, "aliases", [])
            hidden = getattr(cmd_func, "hidden", False)

            # map command name and aliases to the function
            self.cmd_map[cmd_name] = cmd_func
            self.alias_to_cmd_name_map[cmd_name] = cmd_name
            for i in aliases:
                self.cmd_map[i] = cmd_func
                self.alias_to_cmd_name_map[i] = cmd_name

            if cmd_func.__doc__:
                # split doctext into lines, allow the first line to start at a new line
                help_lines = cmd_func.__doc__.lstrip().splitlines()

                # use the first line as help text
                help_text = help_lines.pop(0)

                # use the remaining lines as description
                help_lines = self._remove_leading_spaces_from_lines(help_lines)
                help_desc = "\n".join(help_lines)
                help_desc = help_desc.strip()
            else:
                help_text = ""
                help_desc = ""

            if hidden:
                help_text = argparse.SUPPRESS

            subparser = self.subparsers.add_parser(
                cmd_name,
                aliases=aliases,
                help=help_text,
                description=help_desc,
                prog=self.get_subcommand_prog(cmd_name),
                formatter_class=HelpFormatter,
                conflict_handler="resolve",
            )

            # add hidden copy of global options so they can be used in any place
            self.add_global_options(subparser, suppress=True)

            # add sub-command options, overriding hidden copies of global options if needed (due to conflict_handler="resolve")
            for option_args, option_kwargs in options:
                subparser.add_argument(*option_args, **option_kwargs)

    def argparse_error(self, *args, **kwargs):
        """
        Raise an argument parser error.
        Automatically pick the right parser for the main program or a subcommand.
        """
        if not self.options.command:
            parser = self.argparser
        else:
            parser = self.subparsers._name_parser_map.get(self.options.command, self.argparser)
        parser.error(*args, **kwargs)

    def pre_argparse(self):
        """
        Hook method executed after `.main()` creates `.argparser` instance
        and before `parse_args()` is called.
        """
        pass

    def add_global_options(self, parser, suppress=False):
        """
        Add options to the main argument parser and all subparsers.
        """
        pass

    def post_argparse(self):
        """
        Hook method executed after `.main()` calls `parse_args()`.
        When called, `.options` and `.args` hold the results of `parse_args()`.
        """
        pass

    def main(self, argv=None):
        if argv is None:
            argv = sys.argv
        else:
            argv = argv[:]  # don't modify caller's list

        self.create_argparser()

        self.options, self.args = self.argparser.parse_known_args(argv[1:])
        unrecognized = [i for i in self.args if i.startswith("-")]
        if unrecognized:
            self.argparser.error(f"unrecognized arguments: {' '.join(unrecognized)}")

        self.post_argparse()

        if not self.options.command:
            self.argparser.error("Please specify a command")

        # find the `do_*` function to call by its name
        cmd = self.cmd_map[self.options.command]
        # run the command with parsed args

        sig = inspect.signature(cmd)
        arg_names = list(sig.parameters.keys())
        if arg_names == ["subcmd", "opts"]:
            # positional args specified manually via @cmdln.option
            if self.args:
                self.argparser.error(f"unrecognized arguments: {' '.join(self.args)}")
            cmd(self.options.command, self.options)
        elif arg_names == ["subcmd", "opts", "args"]:
            # positional args are the remaining (unrecognized) args
            cmd(self.options.command, self.options, *self.args)
        else:
            # positional args are the remaining (unrecongnized) args
            # and the do_* handler takes other arguments than "subcmd", "opts", "args"
            import warnings
            warnings.warn(
                f"do_{self.options.command}() handler has deprecated signature. "
                f"It takes the following args: {arg_names}, while it should be taking ['subcmd', 'opts'] "
                f"and handling positional arguments explicitly via @cmdln.option.",
                FutureWarning
            )
            try:
                cmd(self.options.command, self.options, *self.args)
            except TypeError as e:
                if e.args[0].startswith("do_"):
                    sys.exit(str(e))
                raise

    @alias("?")
    def do_help(self, subcmd, opts, *args):
        """
        Give detailed help on a specific sub-command

        usage:
          %(prog)s [SUBCOMMAND]
        """
        if not args:
            self.argparser.print_help()
            return

        for action in self.argparser._actions:
            if not isinstance(action, argparse._SubParsersAction):
                continue

            for choice, subparser in action.choices.items():
                if choice == args[0]:
                    subparser.print_help()
                    return



import os
import unittest

import osc.core
import osc.oscerr

from .common import OscTestCase


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), "fixtures", "packages")


class PackageMock(osc.core.Package):
    def __init__(self, apiurl, project_name, name):
        """
        Let's override __init__ to avoid loading from a working copy.
        """
        self.apiurl = apiurl
        self.prjname = project_name
        self.name = name


class TestPackage(unittest.TestCase):
    def test_eq(self):
        p1 = PackageMock("http://urlA", "projA", "pkgA")
        p2 = PackageMock("http://urlA", "projA", "pkgA")
        self.assertEqual(p1, p2)

        p1 = PackageMock("http://urlA", "projA", "pkgA")
        p2 = PackageMock("http://urlA", "projA", "pkgB")
        self.assertNotEqual(p1, p2)

        p1 = PackageMock("http://urlA", "projA", "pkgA")
        p2 = PackageMock("http://urlA", "projB", "pkgA")
        self.assertNotEqual(p1, p2)

        p1 = PackageMock("http://urlA", "projA", "pkgA")
        p2 = PackageMock("http://urlB", "projA", "pkgA")
        self.assertNotEqual(p1, p2)

    def test_lt(self):
        p1 = PackageMock("http://urlA", "projA", "pkgA")
        p2 = PackageMock("http://urlA", "projA", "pkgA")
        self.assertFalse(p1 < p2)

        p1 = PackageMock("http://urlA", "projA", "pkgA")
        p2 = PackageMock("http://urlA", "projA", "pkgB")
        self.assertTrue(p1 < p2)

        p1 = PackageMock("http://urlA", "projA", "pkgA")
        p2 = PackageMock("http://urlA", "projB", "pkgA")
        self.assertTrue(p1 < p2)

        p1 = PackageMock("http://urlA", "projA", "pkgA")
        p2 = PackageMock("http://urlB", "projA", "pkgA")
        self.assertTrue(p1 < p2)

    def test_hash(self):
        p1 = PackageMock("http://urlA", "projA", "pkgA")
        p2 = PackageMock("http://urlA", "projA", "pkgA")
        packages = set()
        packages.add(p1)
        # the second instance appears to be there because it has the same hash
        # it is ok, because we consider such packages equal
        self.assertIn(p2, packages)


class TestPackageFromPaths(OscTestCase):
    def _get_fixtures_dir(self):
        return FIXTURES_DIR

    def test_package_object_dir(self):
        path = "projectA/pkgA"
        path = os.path.join(self.tmpdir, 'osctest', path)
        pac = osc.core.Package(path)

        self.assertEqual(pac.name, "pkgA")
        self.assertEqual(pac.prjname, "projectA")
        self.assertEqual(pac.apiurl, "http://localhost")
        self.assertEqual(pac.todo, [])

    def test_package_object_file(self):
        path = "projectA/pkgA/pkgA.spec"
        path = os.path.join(self.tmpdir, 'osctest', path)
        pac = osc.core.Package(path)

        self.assertEqual(pac.name, "pkgA")
        self.assertEqual(pac.prjname, "projectA")
        self.assertEqual(pac.apiurl, "http://localhost")
        self.assertEqual(pac.todo, ["pkgA.spec"])

    def test_package_object_file_missing(self):
        path = "projectA/pkgA/missing-file"
        path = os.path.join(self.tmpdir, 'osctest', path)
        pac = osc.core.Package(path)

        self.assertEqual(pac.name, "pkgA")
        self.assertEqual(pac.prjname, "projectA")
        self.assertEqual(pac.apiurl, "http://localhost")
        self.assertEqual(pac.todo, ["missing-file"])

    def test_single_package(self):
        paths = ["projectA/pkgA"]
        paths = [os.path.join(self.tmpdir, 'osctest', i) for i in paths]
        pacs = osc.core.Package.from_paths(paths)
        self.assertEqual(len(pacs), 1)

        pac = pacs[0]
        self.assertEqual(pac.name, "pkgA")
        self.assertEqual(pac.prjname, "projectA")
        self.assertEqual(pac.apiurl, "http://localhost")

    def test_duplicates(self):
        # passing a path twice is ok
        paths = ["projectA/pkgA", "projectA/pkgA"]
        paths = [os.path.join(self.tmpdir, 'osctest', i) for i in paths]
        pacs = osc.core.Package.from_paths(paths)
        pac = pacs[0]
        self.assertEqual(pac.name, "pkgA")
        self.assertEqual(pac.prjname, "projectA")
        self.assertEqual(pac.apiurl, "http://localhost")

        # the same package in 2 paths is an error
        paths = ["projectA/pkgA", "projectA/pkgA-symlink"]
        paths = [os.path.join(self.tmpdir, 'osctest', i) for i in paths]
        self.assertRaises(osc.oscerr.PackageExists, osc.core.Package.from_paths, paths)

    def test_one_package_two_files(self):
        paths = ["projectA/pkgA/pkgA.spec", "projectA/pkgA/pkgA.changes"]
        paths = [os.path.join(self.tmpdir, 'osctest', i) for i in paths]
        pacs = osc.core.Package.from_paths(paths)
        self.assertEqual(len(pacs), 1)

        pac = pacs[0]
        self.assertEqual(pac.name, "pkgA")
        self.assertEqual(pac.prjname, "projectA")
        self.assertEqual(pac.apiurl, "http://localhost")

    def test_two_packages(self):
        paths = ["projectA/pkgA", "projectA/pkgB"]
        paths = [os.path.join(self.tmpdir, 'osctest', i) for i in paths]
        pacs = osc.core.Package.from_paths(paths)
        self.assertEqual(len(pacs), 2)

        pac = pacs[0]
        self.assertEqual(pac.name, "pkgA")
        self.assertEqual(pac.prjname, "projectA")
        self.assertEqual(pac.apiurl, "http://localhost")

        pac = pacs[1]
        self.assertEqual(pac.name, "pkgB")
        self.assertEqual(pac.prjname, "projectA")
        self.assertEqual(pac.apiurl, "http://localhost")

    def test_two_projects(self):
        paths = ["projectA/pkgA", "projectA/pkgB", "projectB/pkgA", "projectB/pkgB"]
        paths = [os.path.join(self.tmpdir, 'osctest', i) for i in paths]
        pacs = osc.core.Package.from_paths(paths)
        self.assertEqual(len(pacs), 4)

        pac = pacs[0]
        self.assertEqual(pac.name, "pkgA")
        self.assertEqual(pac.prjname, "projectA")
        self.assertEqual(pac.apiurl, "http://localhost")

        pac = pacs[1]
        self.assertEqual(pac.name, "pkgB")
        self.assertEqual(pac.prjname, "projectA")
        self.assertEqual(pac.apiurl, "http://localhost")

        pac = pacs[2]
        self.assertEqual(pac.name, "pkgA")
        self.assertEqual(pac.prjname, "projectB")
        self.assertEqual(pac.apiurl, "http://localhost")

        pac = pacs[3]
        self.assertEqual(pac.name, "pkgB")
        self.assertEqual(pac.prjname, "projectB")
        self.assertEqual(pac.apiurl, "http://localhost")

    def test_two_apiurls(self):
        paths = ["projectA/pkgA", "projectA/pkgB", "projectA-different-apiurl/pkgA", "projectA-different-apiurl/pkgB"]
        paths = [os.path.join(self.tmpdir, 'osctest', i) for i in paths]
        pacs = osc.core.Package.from_paths(paths)
        self.assertEqual(len(pacs), 4)

        pac = pacs[0]
        self.assertEqual(pac.name, "pkgA")
        self.assertEqual(pac.prjname, "projectA")
        self.assertEqual(pac.apiurl, "http://localhost")

        pac = pacs[1]
        self.assertEqual(pac.name, "pkgB")
        self.assertEqual(pac.prjname, "projectA")
        self.assertEqual(pac.apiurl, "http://localhost")

        pac = pacs[2]
        self.assertEqual(pac.name, "pkgA")
        self.assertEqual(pac.prjname, "projectA")
        self.assertEqual(pac.apiurl, "http://example.com")

        pac = pacs[3]
        self.assertEqual(pac.name, "pkgB")
        self.assertEqual(pac.prjname, "projectA")
        self.assertEqual(pac.apiurl, "http://example.com")

    def test_invalid_package(self):
        paths = ["projectA/pkgA", "projectA"]
        paths = [os.path.join(self.tmpdir, 'osctest', i) for i in paths]
        self.assertRaises(osc.oscerr.NoWorkingCopy, osc.core.Package.from_paths, paths)

    def test_nofail(self):
        # valid package, invalid package, nonexistent package
        paths = ["projectA/pkgA", "projectA", "does-not-exist"]
        paths = [os.path.join(self.tmpdir, 'osctest', i) for i in paths]
        pacs, nopacs = osc.core.Package.from_paths_nofail(paths)

        self.assertEqual(len(pacs), 1)

        pac = pacs[0]
        self.assertEqual(pac.name, "pkgA")
        self.assertEqual(pac.prjname, "projectA")
        self.assertEqual(pac.apiurl, "http://localhost")

        expected = [
            os.path.join(self.tmpdir, "osctest", "projectA"),
            os.path.join(self.tmpdir, "osctest", "does-not-exist"),
        ]
        self.assertEqual(nopacs, expected)


if __name__ == "__main__":
    unittest.main()



import functools

from .. import oscerr
from . import api


@functools.total_ordering
class PackageBase:
    def __init__(self, apiurl, project, package):
        self.apiurl = apiurl
        self.project = project
        self.name = package

        self.rev = None
        self.vrev = None
        self.srcmd5 = None

        self.linkinfo = None
        self.files = []
        directory_node = self._get_directory_node()
        self._load_from_directory_node(directory_node)
        self._meta_node = None

    def __str__(self):
        return f"{self.project}/{self.name}"

    def __repr__(self):
        return super().__repr__() + f"({self})"

    def __hash__(self):
        return hash((self.name, self.project, self.apiurl))

    def __eq__(self, other):
        return (self.name, self.project, self.apiurl) == (other.name, other.project, other.apiurl)

    def __lt__(self, other):
        return (self.name, self.project, self.apiurl) < (other.name, other.project, other.apiurl)

    def _get_directory_node(self):
        raise NotImplementedError

    def _load_from_directory_node(self, directory_node):
        from .. import core as osc_core

        # attributes
        self.rev = directory_node.get("rev")
        self.vrev = directory_node.get("vrev")
        self.srcmd5 = directory_node.get("srcmd5")

        # files
        file_nodes = api.find_nodes(directory_node, "directory", "entry")
        for file_node in file_nodes:
            self.files.append(osc_core.File.from_xml_node(file_node))

        # linkinfo
        linkinfo_node = api.find_node(directory_node, "directory", "linkinfo")
        if linkinfo_node is not None:
            self.linkinfo = osc_core.Linkinfo()
            self.linkinfo.read(linkinfo_node)
            if self.linkinfo.project and not self.linkinfo.package:
                # if the link points to a package with the same name,
                # the name is omitted and we want it present for overall sanity
                self.linkinfo.package = self.name

    def _get_meta_node(self):
        raise NotImplementedError()

    def get_meta_value(self, option):
        if self._meta_node is None or len(self._meta_node) == 0:
            self._meta_node = self._get_meta_node()
        if self._meta_node is None or len(self._meta_node) == 0:
            return None
        node = api.find_node(self._meta_node, "package", option)
        if node is None or not node.text:
            raise oscerr.APIError(f"Couldn't get '{option}' from package _meta")
        return node.text


class ApiPackage(PackageBase):
    def __init__(self, apiurl, project, package, rev=None):
        # for loading the directory node from the API
        # the actual revision is loaded from the directory node
        self.__rev = rev
        super().__init__(apiurl, project, package)

    def _get_directory_node(self):
        url_path = ["source", self.project, self.name]
        url_query = {}
        if self.__rev:
            url_query["rev"] = self.__rev
        return api.get(self.apiurl, url_path, url_query)

    def _get_meta_node(self):
        url_path = ["source", self.project, self.name, "_meta"]
        url_query = {}
        return api.get(self.apiurl, url_path, url_query)


class LocalPackage(PackageBase):
    def __init__(self, path):
        from .. import store as osc_store

        self.dir = path
        self.store = osc_store.Store(self.dir)
        super().__init__(self.store.apiurl, self.store.project, self.store.package)

    def _get_directory_node(self):
        return self.store.read_xml_node("_files", "directory").getroot()

    def _get_meta_node(self):
        return self.store._meta_node



import urllib.error

from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import
from .package import Package
from .package_sources import PackageSources
from .request_action_acceptinfo import RequestActionAcceptinfo
from .request_action_group import RequestActionGroup
from .request_action_grouped import RequestActionGrouped
from .request_action_options import RequestActionOptions
from .request_action_person import RequestActionPerson
from .request_action_source import RequestActionSource
from .request_action_target import RequestActionTarget
from .request_sourcediff import RequestSourcediff


class RequestAction(XmlModel):
    XML_TAG = "action"

    class TypeEnum(str, Enum):
        SUBMIT = "submit"
        DELETE = "delete"
        CHANGE_DEVEL = "change_devel"
        ADD_ROLE = "add_role"
        SET_BUGOWNER = "set_bugowner"
        MAINTENANCE_INCIDENT = "maintenance_incident"
        MAINTENANCE_RELEASE = "maintenance_release"
        RELEASE = "release"
        GROUP = "group"

    type: TypeEnum = Field(
        xml_attribute=True,
    )

    source: Optional[RequestActionSource] = Field(
    )

    target: Optional[RequestActionTarget] = Field(
    )

    person: Optional[RequestActionPerson] = Field(
    )

    group: Optional[RequestActionGroup] = Field(
    )

    grouped_list: Optional[List[RequestActionGrouped]] = Field(
        xml_name="grouped",
    )

    options: Optional[RequestActionOptions] = Field(
    )

    sourcediff: Optional[RequestSourcediff] = Field(
    )

    acceptinfo: Optional[RequestActionAcceptinfo] = Field(
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._allow_new_attributes = True
        # source and target always come from ``self._apiurl`` while devel and factory projects may live elsewhere
        self._devel_apiurl = self._apiurl
        self._factory_apiurl = self._apiurl
        self._factory_project = "openSUSE:Factory"
        self._props = {}
        self._allow_new_attributes = False

    def _get_package(self, package_type):
        key = f"{package_type}_package"
        if key not in self._props:
            func = getattr(self, f"_get_{package_type}_apiurl_project_package")
            apiurl, project, package = func()
            if apiurl is None:
                self._props[key] = None
            else:
                try:
                    self._props[key] = Package.from_api(apiurl, project, package)
                except urllib.error.HTTPError as e:
                    if e.code != 404:
                        raise
                    self._props[key] = None
        return self._props[key]

    def _get_package_sources(self, package_type, *, rev=None):
        key = f"{package_type}_package_sources"
        if key not in self._props:
            func = getattr(self, f"_get_{package_type}_apiurl_project_package")
            apiurl, project, package = func()
            if apiurl is None:
                self._props[key] = None
            else:
                try:
                    self._props[key] = PackageSources.from_api(apiurl, project, package, rev=rev)
                except urllib.error.HTTPError as e:
                    if e.code != 404:
                        raise
                    self._props[key] = None
        return self._props[key]

    def _get_source_apiurl_project_package(self):
        return self._apiurl, self.source.project, self.source.package

    @property
    def source_package(self) -> Optional[Package]:
        """
        Return a ``Package`` object that encapsulates metadata of the source package.
        """
        return self._get_package("source")

    @property
    def source_package_sources(self) -> Optional[PackageSources]:
        """
        Return a ``PackageSources`` object that contains information about the ``source.rev`` revision of the source package sources in OBS SCM.
        """
        if self.source is None:
            return None
        return self._get_package_sources("source", rev=self.source.rev)

    def _get_target_apiurl_project_package(self):
        if self.target is None:
            return None, None, None
        target_project, target_package = self.get_actual_target_project_package()
        return self._apiurl, target_project, target_package

    @property
    def target_package(self) -> Optional[Package]:
        """
        Return a ``Package`` object that encapsulates metadata of the target package.
        """
        return self._get_package("target")

    @property
    def target_package_sources(self) -> Optional[PackageSources]:
        """
        Return a ``PackageSources`` object that contains information about the current revision of the target package sources in OBS SCM.
        """
        return self._get_package_sources("target")

    def _get_factory_apiurl_project_package(self):
        if self.target is None:
            # a new package was submitted, it doesn't exist on target; let's read the package name from the source
            target_project, target_package = None, self.source.package
        else:
            target_project, target_package = self.get_actual_target_project_package()

        if (self._apiurl, target_project) == (self._factory_apiurl, self._factory_project):
            # factory package equals the target package
            return None, None, None

        return self._factory_apiurl, self._factory_project, target_package

    @property
    def factory_package(self) -> Optional[Package]:
        """
        Return a ``Package`` object that encapsulates metadata of the package in the factory project.
        The name of the package equals the target package name.
        """
        return self._get_package("factory")

    @property
    def factory_package_sources(self) -> Optional[PackageSources]:
        """
        Return a ``PackageSources`` object that contains information about the current revision of the factory package sources in OBS SCM.
        """
        return self._get_package_sources("factory")

    def _get_devel_apiurl_project_package(self):
        if self.factory_package is None:
            return None, None, None

        devel = self.factory_package.devel
        if devel is None:
            return None, None, None

        return (
            self._devel_apiurl,
            devel.project,
            devel.package or self.factory_package.name,
        )

    @property
    def devel_package(self) -> Optional[Package]:
        """
        Return a ``Package`` object that encapsulates metadata of the package in the devel project.
        The devel project name and package name come from ``self.factory_package.devel``.
        If the devel package name is not set, target package name is used.
        """
        return self._get_package("devel")

    @property
    def devel_package_sources(self) -> Optional[PackageSources]:
        """
        Return a ``PackageSources`` object that contains information about the current revision of the devel package sources in OBS SCM.
        """
        return self._get_package_sources("devel")

    def get_actual_target_project_package(self) -> Tuple[str, str]:
        """
        Return the target project and package names because maintenance incidents require special handling.

        The target project for maintenance incidents is virtual and cannot be queried.
        The actual target project is specified in target's ``releaseproject`` field.

        Also the target package for maintenance incidents is not set explicitly.
        It is extracted from ``releasename`` field from the source metadata.
        If ``releasename`` is not defined, source package name is used.
        """

        if self.type == "maintenance_incident":
            # dmach's note on security:
            # The ``releaseproject`` is baked into the target information in the request and that's perfectly ok.
            # The ``releasename`` is part of the source package metadata and *may change* after the request is created.
            # After consulting this with OBS developers, I believe this doesn't represent any security issue
            # because the project is fixed and tampering with ``releasename`` might only lead to inconsistent naming,
            # the package would still end up it the same project.

            # target.releaseproject is always set for a maintenance_incident
            assert self.target
            assert self.target.releaseproject
            project = self.target.releaseproject

            # the target package is not specified
            # we need to extract it from source package's metadata or use source package name as a fallback
            assert self.source_package
            if self.source_package.releasename:
                package = self.source_package.releasename.split(".")[0]
            else:
                package = self.source_package.name

            return project, package

        assert self.target
        assert self.target.project
        assert self.target.package
        return self.target.project, self.target.package



import difflib

import osc.commandline
from .. import oscerr


class RepoAddCommand(osc.commandline.OscCommand):
    """
    Add a repository to project meta
    """

    name = "add"
    parent = "RepoCommand"

    def init_arguments(self):
        self.add_argument(
            "project",
            help="Name of the project",
        )
        self.add_argument(
            "--repo",
            metavar="NAME",
            required=True,
            help="Name of the repository we're adding",
        )
        self.add_argument(
            "--arch",
            dest="arches",
            metavar="[ARCH]",
            action="append",
            required=True,
            help="Architecture of the repository. Can be specified multiple times.",
        )
        self.add_argument(
            "--path",
            dest="paths",
            metavar="[PROJECT/REPO]",
            action="append",
            required=True,
            help="Path associated to the repository. Format is PROJECT/REPO. Can be specified multiple times.",
        )
        self.add_argument(
            "--disable-publish",
            action="store_true",
            default=False,
            help="Disable publishing the added repository",
        )
        self.add_argument(
            "--yes",
            action="store_true",
            help="Proceed without asking",
        )

    def run(self, args):
        from .. import obs_api
        from ..output import get_user_input

        paths = []
        for path in args.paths:
            if "/" not in path:
                self.parser.error(f"Invalid path (expected format is PROJECT/REPO): {path}")
            project, repo = path.split("/")
            paths.append({"project": project, "repository": repo})

        project_obj = obs_api.Project.from_api(args.apiurl, args.project)
        old = project_obj.to_string()

        matching_repos = [i for i in project_obj.repository_list or [] if i.name == args.repo]
        if matching_repos:
            raise oscerr.OscValueError(f"Repository '{args.repo}' already exists in project meta")

        project_obj.repository_list.append(
            {
                "name": args.repo,
                "arch_list": args.arches,
                "path_list": paths,
            }
        )

        if args.disable_publish:
            matching_publish_disable_repos = [
                i for i in project_obj.publish_list or [] if i.flag == "disable" and i.repository == args.repo
            ]
            if not matching_publish_disable_repos:
                if project_obj.publish_list is None:
                    project_obj.publish_list = []
                project_obj.publish_list.append(
                    {
                        "flag": "disable",
                        "repository": args.repo,
                    }
                )

        if not args.yes:
            new = project_obj.to_string()
            diff = difflib.unified_diff(old.splitlines(), new.splitlines(), fromfile="old", tofile="new")
            print("\n".join(diff))
            print()

            reply = get_user_input(
                f"""
                You're changing meta of project '{args.project}'
                Do you want to apply the changes?
                """,
                answers={"y": "yes", "n": "no"},
            )

            if reply == "n":
                raise oscerr.UserAbort()

        project_obj.to_api(args.apiurl)



import sys

import osc.commandline_git


class LoginUpdateCommand(osc.commandline_git.GitObsCommand):
    """
    Update a Gitea credentials entry
    """

    name = "update"
    parent = "LoginCommand"

    def init_arguments(self):
        self.parser.add_argument("name")
        self.parser.add_argument("--new-name")
        self.parser.add_argument("--new-url")
        self.parser.add_argument("--new-user")
        self.parser.add_argument("--new-token")
        self.parser.add_argument("--new-ssh-key")
        self.parser.add_argument("--set-as-default", action="store_true")

    def run(self, args):
        print(f"Updating a Gitea credentials entry with name '{args.name}' ...", file=sys.stderr)
        print(f" * Config path: {self.gitea_conf.path}", file=sys.stderr)
        print("", file=sys.stderr)

        # TODO: try to authenticate to verify that the updated entry works

        original_login = self.gitea_conf.get_login(args.name)
        print("Original entry:")
        print(original_login.to_human_readable_string())

        updated_login = self.gitea_conf.update_login(
            args.name,
            new_name=args.new_name,
            new_url=args.new_url,
            new_user=args.new_user,
            new_token=args.new_token,
            new_ssh_key=args.new_ssh_key,
            set_as_default=args.set_as_default,
        )
        print("")
        print("Updated entry:")
        print(updated_login.to_human_readable_string())



# Copyright 2009 Marcus Huewe <suse-tux@gmx.de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation;
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


import os
import re
import stat
import sys
from io import BytesIO
from typing import Union


class ArError(Exception):
    """Base class for all ar related errors"""

    def __init__(self, fn: bytes, msg: str):
        super().__init__()
        self.file = fn
        self.msg = msg

    def __str__(self):
        return f"{self.msg}: {self.file.decode('utf-8')}"


class ArHdr:
    """Represents an ar header entry"""

    def __init__(self, fn: bytes, date: bytes, uid: bytes, gid: bytes, mode: bytes, size: bytes, fmag: bytes, off: bytes):
        self.file = fn.strip()
        self.date = date.strip()
        self.uid = uid.strip()
        self.gid = gid.strip()
        if not mode.strip():
            # provide a dummy mode for the ext_fn hdr
            mode = b"0"
        self.mode = stat.S_IMODE(int(mode, 8))
        self.size = int(size)
        self.fmag = fmag
        # data section starts at off and ends at off + size
        self.dataoff = int(off)

    def __str__(self):
        return '%16s %d' % (self.file, self.size)


class ArFile(BytesIO):
    """Represents a file which resides in the archive"""

    def __init__(self, fn, uid, gid, mode, buf):
        super().__init__(buf)
        self.name = fn
        self.uid = uid
        self.gid = gid
        self.mode = mode

    def saveTo(self, dir=None):
        """
        writes file to dir/filename if dir isn't specified the current
        working dir is used. Additionally it tries to set the owner/group
        and permissions.
        """
        if self.name.startswith(b"/"):
            raise ArError(self.name, "Extracting files with absolute paths is not supported for security reasons")

        if not dir:
            dir = os.getcwdb()
        fn = os.path.join(dir, self.name.decode("utf-8"))

        dir_path, _ = os.path.split(fn)
        if dir_path:
            os.makedirs(dir_path, exist_ok=True)

        with open(fn, 'wb') as f:
            f.write(self.getvalue())
        os.chmod(fn, self.mode)
        uid = self.uid
        if uid != os.geteuid() or os.geteuid() != 0:
            uid = -1
        gid = self.gid
        if gid not in os.getgroups() or os.getegid() != 0:
            gid = -1
        os.chown(fn, uid, gid)
        return fn

    def __str__(self):
        return '%s %s %s %s' % (self.name, self.uid,
                                self.gid, self.mode)


class Ar:
    """
    Represents an ar archive (only GNU format is supported).
    Readonly access.
    """
    hdr_len = 60
    hdr_pat = re.compile(b'^(.{16})(.{12})(.{6})(.{6})(.{8})(.{10})(.{2})',
                         re.DOTALL)

    def __init__(self, fn=None, fh=None):
        if fn is None and fh is None:
            raise ValueError('either \'fn\' or \'fh\' must be is not None')
        if fh is not None:
            self.__file = fh
            self.__closefile = False
            self.filename = fh.name
        else:
            # file object: will be closed in __del__()
            self.__file = None
            self.__closefile = True
            self.filename = fn
        self._init_datastructs()

    def __del__(self):
        if self.__file and self.__closefile:
            self.__file.close()

    def _init_datastructs(self):
        self.hdrs = []
        self.ext_fnhdr = None

    def _appendHdr(self, hdr):
        # GNU uses an internal '//' file to store very long filenames
        if hdr.file.startswith(b'//'):
            self.ext_fnhdr = hdr
        else:
            self.hdrs.append(hdr)

    def _fixupFilenames(self):
        """
        support the GNU approach for very long filenames:
        every filename which exceeds 16 bytes is stored in the data section of a special file ('//')
        and the filename in the header of this long file specifies the offset in the special file's
        data section. The end of such a filename is indicated with a trailing '/'.
        Another special file is the '/' which contains the symbol lookup table.
        """

        # read extended header with long file names and then only seek into the right offsets
        ext_fnhdr_data = None
        if self.ext_fnhdr:
            self.__file.seek(self.ext_fnhdr.dataoff, os.SEEK_SET)
            ext_fnhdr_data = self.__file.read(self.ext_fnhdr.size)

        for h in self.hdrs:
            if h.file == b'/':
                continue

            if h.file.endswith(b"/"):
                # regular file name
                h.file = h.file[:-1]
                continue

            if not h.file.startswith(b'/'):
                continue

            # long file name
            assert h.file[0:1] == b"/"
            assert ext_fnhdr_data is not None
            start = int(h.file[1:])
            end = ext_fnhdr_data.find(b'/', start)

            if end == -1:
                raise ArError(b'//', 'invalid data section - trailing slash (off: %d)' % start)

            h.file = ext_fnhdr_data[start:end]

    def _get_file(self, hdr):
        self.__file.seek(hdr.dataoff, os.SEEK_SET)
        return ArFile(hdr.file, hdr.uid, hdr.gid, hdr.mode,
                      self.__file.read(hdr.size))

    def read(self):
        """reads in the archive."""
        if not self.__file:
            self.__file = open(self.filename, 'rb')
        else:
            self.__file.seek(0, os.SEEK_SET)
        self._init_datastructs()
        data = self.__file.read(7)
        if data != b'!<arch>':
            raise ArError(self.filename, 'no ar archive')
        pos = 8
        while len(data) != 0:
            self.__file.seek(pos, os.SEEK_SET)
            data = self.__file.read(self.hdr_len)
            if not data:
                break
            pos += self.hdr_len
            m = self.hdr_pat.search(data)
            if not m:
                raise ArError(self.filename, 'unexpected hdr entry')
            args = m.groups() + (pos, )
            hdr = ArHdr(*args)
            self._appendHdr(hdr)
            # data blocks are 2 bytes aligned - if they end on an odd
            # offset ARFMAG[0] will be used for padding (according to the current binutils code)
            pos += hdr.size + (hdr.size & 1)
        self._fixupFilenames()

    def get_file(self, fn: Union[str, bytes]):
        # accept str for better user experience
        if isinstance(fn, str):
            fn = fn.encode("utf-8")
        for h in self.hdrs:
            if h.file == fn:
                return self._get_file(h)
        return None

    def __iter__(self):
        for h in self.hdrs:
            if h.file == b'/':
                continue
            yield self._get_file(h)


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print('usage: %s <arfile>' % sys.argv[0])
        sys.exit(1)
    # a potential user might want to pass a bytes instead of a str
    # to make sure that the ArError's file attribute is always a
    # bytes
    ar = Ar(fn=sys.argv[1])
    ar.read()
    for hdr in ar.hdrs:
        print(hdr)



# Copyright (C) 2008 Novell Inc.  All rights reserved.
# This program is free software; it may be used, copied, modified
# and distributed under the terms of the GNU General Public Licence,
# either version 2, or (at your option) any later version.


class OscBaseError(Exception):
    def __init__(self, args=()):
        super().__init__()
        self.args = args

    def __str__(self):
        return ''.join(self.args)


class UserAbort(OscBaseError):
    """Exception raised when the user requested abortion"""


class ConfigError(OscBaseError):
    """Exception raised when there is an error in the config file"""

    def __init__(self, msg, fname):
        super().__init__()
        self.msg = msg
        self.file = fname

    def __str__(self):
        return f"Error in config file {self.file}\n   {self.msg}"


class ConfigMissingApiurl(ConfigError):
    """Exception raised when a apiurl does not exist in the config file"""

    def __init__(self, msg, fname, url):
        super().__init__(msg, fname)
        self.url = url


class ConfigMissingCredentialsError(ConfigError):
    def __init__(self, msg, fname, url):
        super().__init__(msg, fname)
        self.url = url


class APIError(OscBaseError):
    """Exception raised when there is an error in the output from the API"""

    def __init__(self, msg):
        super().__init__()
        self.msg = msg

    def __str__(self):
        return f"{self.__class__.__name__}: {self.msg}"


class NotFoundAPIError(APIError):
    """
    Requested data was not found.
    """


class NoConfigfile(OscBaseError):
    """Exception raised when osc's configfile cannot be found"""

    def __init__(self, fname, msg):
        super().__init__()
        self.file = fname
        self.msg = msg

    def __str__(self):
        return f"Config file cannot be found: {self.file}\n   {self.msg}"


class ExtRuntimeError(OscBaseError):
    """Exception raised when there is a runtime error of an external tool"""

    def __init__(self, msg, fname):
        super().__init__()
        self.msg = msg
        self.file = fname


class ServiceRuntimeError(OscBaseError):
    """Exception raised when the execution of a source service failed"""

    def __init__(self, msg):
        super().__init__()
        self.msg = msg


class WrongArgs(OscBaseError):
    """Exception raised by the cli for wrong arguments usage"""


class WrongOptions(OscBaseError):
    """Exception raised by the cli for wrong option usage"""
    # def __str__(self):
    #    s = 'Sorry, wrong options.'
    #    if self.args:
    #        s += '\n' + self.args
    #    return s


class NoWorkingCopy(OscBaseError):
    """Exception raised when directory is neither a project dir nor a package dir"""


class NotMissing(OscBaseError):
    """Exception raised when link target should not exist, but it does"""


class WorkingCopyWrongVersion(OscBaseError):
    """Exception raised when working copy's .osc/_osclib_version doesn't match"""


class WorkingCopyOutdated(OscBaseError):
    """Exception raised when the working copy is outdated.
    It takes a tuple with three arguments: path to wc,
    revision that it has, revision that it should have.
    """

    def __str__(self):
        return ('Working copy \'%s\' is out of date (rev %s vs rev %s).\n'
                'Looks as if you need to update it first.'
                % (self.args[0], self.args[1], self.args[2]))


class ProjectError(OscBaseError):
    """Base class for all Project related exceptions"""

    def __init__(self, prj, msg=None):
        super().__init__()
        self.prj = prj
        self.msg = msg

    def __str__(self):
        result = f"{self.__class__.__name__}: {self.prj}"
        if self.msg:
            result += f": {self.msg}"
        return result


class PackageError(OscBaseError):
    """Base class for all Package related exceptions"""

    def __init__(self, prj, pac, msg=None):
        super().__init__()
        self.prj = prj
        self.pac = pac
        self.msg = msg

    def __str__(self):
        result = f"{self.__class__.__name__}: {self.prj}/{self.pac}"
        if self.msg:
            result += f": {self.msg}"
        return result


class WorkingCopyInconsistent(PackageError):
    """Exception raised when the working copy is in an inconsistent state"""

    def __init__(self, prj, pac, dirty_files, msg):
        super().__init__(prj, pac, msg)
        self.dirty_files = dirty_files


class LinkExpandError(PackageError):
    """Exception raised when source link expansion fails"""


class OscIOError(OscBaseError):
    def __init__(self, e, msg):
        super().__init__()
        self.e = e
        self.msg = msg


class OscValueError(OscBaseError):
    """
    Invalid argument value (of correct type).
    """
    pass


class OscInvalidRevision(OscValueError):
    """
    Invalid revision value.
    """

    def __str__(self):
        return f"Invalid revision value: {''.join(self.args)}"


class PackageNotInstalled(OscBaseError):
    """
    Exception raised when a package is not installed on local system
    """

    def __init__(self, pkg):
        super().__init__((pkg,))

    def __str__(self):
        return f'Package {self.args} is required for this operation'


class SignalInterrupt(Exception):
    """Exception raised on SIGTERM and SIGHUP."""


class PackageExists(PackageError):
    """
    Exception raised when a local object already exists
    """


class PackageMissing(PackageError):
    """
    Exception raised when a local object doesn't exist
    """


class PackageFileConflict(PackageError):
    """
    Exception raised when there's a file conflict.
    Conflict doesn't mean an unsuccessfull merge in this context.
    """

    def __init__(self, prj, pac, file, msg):
        super().__init__(prj, pac, msg)
        self.file = file


class PackageInternalError(PackageError):
    pass


# vim: sw=4 et



import os

XDG_DATA_HOME = os.environ.get("XDG_DATA_HOME", "~/.local/share")
XDG_CONFIG_HOME = os.environ.get("XDG_CONFIG_HOME", "~/.config")
XDG_STATE_HOME = os.environ.get("XDG_STATE_HOME", "~/.local/state")
XDG_CACHE_HOME = os.environ.get("XDG_CACHE_HOME", "~/.cache")



from .key_value_table import KeyValueTable
from .input import get_user_input
from .output import get_default_pager
from .output import pipe_to_pager
from .output import print_msg
from .output import run_pager
from .output import sanitize_text
from .output import safe_print
from .output import safe_write
from .tty import colorize
from .widechar import wc_ljust
from .widechar import wc_width



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class RequestActionOptions(XmlModel):
    XML_TAG = "options"

    class SourceupdateEnum(str, Enum):
        UPDATE = "update"
        NOUPDATE = "noupdate"
        CLEANUP = "cleanup"

    sourceupdate: Optional[SourceupdateEnum] = Field(
    )

    class UpdatelinkEnum(str, Enum):
        TRUE = "true"
        FALSE = "false"

    updatelink: Optional[UpdatelinkEnum] = Field(
    )

    class MakeoriginolderEnum(str, Enum):
        TRUE = "true"
        FALSE = "false"

    makeoriginolder: Optional[MakeoriginolderEnum] = Field(
    )



import os
import subprocess


def get_git_archive_version():
    """
    Return version that is set by git during `git archive`.
    The returned format is equal to what `git describe --tags` returns.
    """
    # the `version` variable contents get substituted during `git archive`
    # it requires adding this to .gitattributes: <path to this file> export-subst
    version = "1.12.1"
    if version.startswith(("$", "%")):
        # "$": version hasn't been substituted during `git archive`
        # "%": "Format:" and "$" characters get removed from the version string (a GitHub bug?)
        return None
    return version


def get_git_version():
    """
    Determine version from git repo by calling `git describe --tags`.
    """
    cmd = ["git", "describe", "--tags"]
    # run the command from the place where this file is placed
    # to ensure that we're in a git repo
    cwd = os.path.dirname(__file__)
    try:
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
    except OSError:
        # `git` command not found
        return None
    stdout, _ = proc.communicate()

    if proc.returncode != 0:
        return None

    version = stdout.strip().decode("utf-8")
    return version


def get_version(version):
    """
    Get the most relevant version of the software:
    1. the version set during `git archive`
    2. the version from the git tags by calling `git describe --tags`
    3. the version explicitly specified in the source code

    The version conforms PEP 440.
    """
    # use version from the archive
    git_version = get_git_archive_version()

    # use version from the git repo
    if not git_version:
        git_version = get_git_version()

    # unable to determine version from git
    if not git_version:
        return version

    if "-" not in git_version:
        git_tag = git_version
        git_commits = None
        git_hash = None
    else:
        git_tag, git_commits, git_hash = git_version.rsplit("-", 2)
        git_commits = int(git_commits)
        # remove the 'g' prefix from hash
        git_hash = git_hash[1:]

    # removing "~" because it is not an allowed character in git tags
    # and also because the normalized form is (for example) 1.0.0b0
    if version and git_tag != version.replace("~", ""):
        # Git tag doesn't correspond with version specified in the source code.
        # The most common reason is that forks do not have their tags in sync with upstream.
        # In that case just return the version specified in the source code.
        return version

    result = git_tag
    if git_hash:
        result += f"+{git_commits}.git.{git_hash}"

    return result



import unittest

from osc.core import Request


class TestRequest(unittest.TestCase):
    def test_eq(self):
        req1 = Request()
        req1.reqid = 1
        req2 = Request()
        req2.reqid = 1
        self.assertEqual(req1, req2)

    def test_lt(self):
        req1 = Request()
        req1.reqid = 1
        req2 = Request()
        req2.reqid = 2
        self.assertTrue(req1 < req2)

    def test_gt(self):
        req1 = Request()
        req1.reqid = 2
        req2 = Request()
        req2.reqid = 1
        self.assertTrue(req1 > req2)

    def test_sort(self):
        req1 = Request()
        req1.reqid = 2
        req2 = Request()
        req2.reqid = 1
        requests = [req1, req2]
        requests.sort()
        self.assertEqual(requests[0].reqid, 1)
        self.assertEqual(requests[1].reqid, 2)


if __name__ == "__main__":
    unittest.main()



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import
from .linkinfo import Linkinfo
from .package_sources_file import PackageSourcesFile
from .serviceinfo import Serviceinfo


class PackageSources(XmlModel):
    XML_TAG = "directory"

    name: str = Field(
        xml_attribute=True,
    )

    rev: str = Field(
        xml_attribute=True,
    )

    vrev: Optional[str] = Field(
        xml_attribute=True,
    )

    srcmd5: str = Field(
        xml_attribute=True,
    )

    linkinfo: Optional[Linkinfo] = Field(
    )

    serviceinfo: Optional[Serviceinfo] = Field(
    )

    file_list: Optional[List[PackageSourcesFile]] = Field(
        xml_name="entry",
    )

    @classmethod
    def from_api(
        cls,
        apiurl: str,
        project: str,
        package: str,
        *,
        deleted: Optional[bool] = None,
        expand: Optional[bool] = None,
        meta: Optional[bool] = None,
        rev: Optional[str] = None,
    ):
        """
        :param deleted: Set to ``True`` to list source files of a deleted package.
                        Throws 400: Bad Request if such package exists.
        :param expand: Expand links.
        :param meta: Set to ``True`` to list metadata file (``_meta``) instead of the sources.
        :param rev: Show sources of the specified revision.
        """
        from ..core import revision_is_empty

        if revision_is_empty(rev):
            rev = None

        url_path = ["source", project, package]
        url_query = {
            "deleted": deleted,
            "expand": expand,
            "meta": meta,
            "rev": rev,
        }
        response = cls.xml_request("GET", apiurl, url_path, url_query)
        return cls.from_file(response, apiurl=apiurl)



import os
import sys


try:
    IS_INTERACTIVE = os.isatty(sys.stdout.fileno())
except OSError:
    IS_INTERACTIVE = False


ESCAPE_CODES = {
    "reset": "\033[0m",
    "bold": "\033[1m",
    "dim": "\033[2m",
    "italic": "\033[3m",
    "underline": "\033[4m",
    "blink": "\033[5m",
    "black": "\033[30m",
    "red": "\033[31m",
    "green": "\033[32m",
    "yellow": "\033[33m",
    "blue": "\033[34m",
    "magenta": "\033[35m",
    "cyan": "\033[36m",
    "white": "\033[37m",
    "bg_black": "\033[40m",
    "bg_red": "\033[41m",
    "bg_green": "\033[42m",
    "bg_yellow": "\033[43m",
    "bg_blue": "\033[44m",
    "bg_magenta": "\033[45m",
    "bg_cyan": "\033[46m",
    "bg_white": "\033[47m",
}


def colorize(text, color):
    """
    Colorize `text` if the `color` is specified and we're running in an interactive terminal.
    """
    if not IS_INTERACTIVE:
        return text

    if not color:
        return text

    if not text:
        return text

    result = ""
    for i in color.split(","):
        result += ESCAPE_CODES[i]
    result += text
    result += ESCAPE_CODES["reset"]
    return result



import sys

import osc.commandline_git


class PullRequestSearchCommand(osc.commandline_git.GitObsCommand):
    """
    Search pull requests in the whole gitea instance
    """

    name = "search"
    parent = "PullRequestCommand"

    def init_arguments(self):
        self.add_argument(
            "--state",
            choices=["open", "closed"],
            default="open",
            help="Filter by state: open, closed (default: open)",
        )
        self.add_argument(
            "--title",
            help="Filter by substring in title",
        )
        self.add_argument(
            "--owner",
            help="Filter by owner of the repository associated with the pull requests",
        )
        self.add_argument(
            "--label",
            dest="labels",
            metavar="LABEL",
            action="append",
            help="Filter by associated labels. Non existent labels are discarded. Can be specified multiple times.",
        )
        self.add_argument(
            "--assigned",
            action="store_true",
            help="Filter pull requests assigned to you",
        )
        self.add_argument(
            "--created",
            action="store_true",
            help="Filter pull requests created by you",
        )
        self.add_argument(
            "--mentioned",
            action="store_true",
            help="Filter pull requests mentioning you",
        )
        self.add_argument(
            "--review-requested",
            action="store_true",
            help="Filter pull requests requesting your review",
        )

    def run(self, args):
        from osc import gitea_api

        self.print_gitea_settings()

        data = gitea_api.PullRequest.search(
            self.gitea_conn,
            state=args.state,
            title=args.title,
            owner=args.owner,
            labels=args.labels,
            assigned=args.assigned,
            created=args.created,
            mentioned=args.mentioned,
            review_requested=args.review_requested,
        ).json()

        text = gitea_api.PullRequest.list_to_human_readable_string(data, sort=True)
        if text:
            print(text)
            print("", file=sys.stderr)

        print(f"Total entries: {len(data)}", file=sys.stderr)



import textwrap

from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import
from .status import Status


class Token(XmlModel):
    XML_TAG = "entry"

    id: int = Field(
        xml_attribute=True,
        description=textwrap.dedent(
            """
            The unique id of this token.
            """
        ),
    )

    string: str = Field(
        xml_attribute=True,
        description=textwrap.dedent(
            """
            The token secret. This string can be used instead of the password to
            authenticate the user or to trigger service runs via the
            `POST /trigger/runservice` route.
            """
        ),
    )

    description: Optional[str] = Field(
        xml_attribute=True,
        description=textwrap.dedent(
            """
            This attribute can be used to identify a token from the list of tokens
            of a user.
            """
        ),
    )

    project: Optional[str] = Field(
        xml_attribute=True,
        description=textwrap.dedent(
            """
            If this token is bound to a specific package, then the packages'
            project is available in this attribute.
            """
        ),
    )

    package: Optional[str] = Field(
        xml_attribute=True,
        description=textwrap.dedent(
            """
            The package name to which this token is bound, if it has been created
            for a specific package. Otherwise this attribute and the project
            attribute are omitted.
            """
        ),
    )

    class Kind(str, Enum):
        RSS = "rss"
        REBUILD = "rebuild"
        RELEASE = "release"
        RUNSERVICE = "runservice"
        WIPE = "wipe"
        WORKFLOW = "workflow"

    kind: Kind = Field(
        xml_attribute=True,
        description=textwrap.dedent(
            """
            This attribute specifies which actions can be performed via this token.
            - rss: used to retrieve the notification RSS feed
            - rebuild: trigger rebuilds of packages
            - release: trigger project releases
            - runservice: run a service via the POST /trigger/runservice route
            - wipe: trigger wipe of binary artifacts
            - workflow: trigger SCM/CI workflows, see https://openbuildservice.org/help/manuals/obs-user-guide/cha.obs.scm_ci_workflow_integration.html
            """
        ),
    )

    triggered_at: str = Field(
        xml_attribute=True,
        description=textwrap.dedent(
            """
            The date and time a token got triggered the last time.
            """
        ),
    )

    def to_human_readable_string(self) -> str:
        """
        Render the object as a human readable string.
        """
        from ..output import KeyValueTable

        table = KeyValueTable()
        table.add("ID", str(self.id))
        table.add("String", self.string, color="bold")
        table.add("Operation", self.kind)
        table.add("Description", self.description)
        table.add("Project", self.project)
        table.add("Package", self.package)
        table.add("Triggered at", self.triggered_at)
        return f"{table}"

    @classmethod
    def do_list(cls, apiurl: str, user: str):
        from ..util.xml import xml_parse

        url_path = ["person", user, "token"]
        url_query = {}
        response = cls.xml_request("GET", apiurl, url_path, url_query)
        root = xml_parse(response).getroot()
        assert root.tag == "directory"
        result = []
        for node in root:
            result.append(cls.from_xml(node, apiurl=apiurl))
        return result

    @classmethod
    def cmd_create(
        cls,
        apiurl: str,
        user: str,
        *,
        operation: Optional[str] = None,
        project: Optional[str] = None,
        package: Optional[str] = None,
        scm_token: Optional[str] = None,
    ):
        if operation == "workflow" and not scm_token:
            raise ValueError('``operation`` = "workflow" requires ``scm_token``')

        url_path = ["person", user, "token"]
        url_query = {
            "cmd": "create",
            "operation": operation,
            "project": project,
            "package": package,
            "scm_token": scm_token,
        }
        response = cls.xml_request("POST", apiurl, url_path, url_query)
        return Status.from_file(response, apiurl=apiurl)

    @classmethod
    def do_delete(cls, apiurl: str, user: str, token: str):
        url_path = ["person", user, "token", token]
        url_query = {}
        response = cls.xml_request("DELETE", apiurl, url_path, url_query)
        return Status.from_file(response, apiurl=apiurl)

    @classmethod
    def do_trigger(
        cls,
        apiurl: str,
        token: str,
        *,
        operation: Optional[str] = None,
        project: Optional[str] = None,
        package: Optional[str] = None,
        repo: Optional[str] = None,
        arch: Optional[str] = None,
        target_project: Optional[str] = None,
        target_repo: Optional[str] = None,
        set_release: Optional[str] = None,
    ):
        if operation:
            url_path = ["trigger", operation]
        else:
            url_path = ["trigger"]

        url_query = {
            "project": project,
            "package": package,
            "repository": repo,
            "architecture": arch,
            "targetproject": target_project,
            "targetrepository": target_repo,
            "setrelease": set_release,
        }

        headers = {
            "Content-Type": "application/octet-stream",
            "Authorization": f"Token {token}",
        }

        response = cls.xml_request("POST", apiurl, url_path, url_query, headers=headers)
        return Status.from_file(response, apiurl=apiurl)



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class PersonWatchlistRequest(XmlModel):
    XML_TAG = "request"

    number: str = Field(
        xml_attribute=True,
    )
